import ANNarchy.parser.report.LatexParser as LatexParser
from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.GlobalObjects import GlobalObjectManager
from ANNarchy.intern import Messages
from ANNarchy.extensions.bold.BoldModel import BoldModel

from ANNarchy.inputs import input_type_list, input_name_list
from ANNarchy.core.Synapse import Synapse
from ANNarchy.core.Neuron import Neuron

import numpy as np
import os

##################################
### Templates
##################################

header = r"""
%  LaTeX file for generating the Model Description Table in Fig. 5 of
%
%  Nordlie E, Gewaltig M-O, Plesser HE (2009)
%  Towards Reproducible Descriptions of Neuronal Network Models.
%  PLoS Comput Biol 5(8): e1000456.
%
%  Paper URL : http://dx.doi.org/10.1371/journal.pcbi.1000456
%  Figure URL: http://dx.doi.org/10.1371/journal.pcbi.1000456.g005
%
%  This file is released under a
%
%   Creative Commons Attribution, non-commercial, share-alike licence
%   http://creativecommons.org/licenses/by-nc-sa/3.0/de/deed.en
%
%  with the following specifications:
%
%  1. When publishing tables generated from this LaTeX file and modified
%     versions of it, you must cite the paper by Nordlie et al given above.
%
%  2. The non-commercial clause applies only to the distribution of THIS FILE
%     and LaTeX source code files derived from it. You may commercially publish
%     documents generated using this file and derivatived versions of this file.
%
%  Contact: Hans Ekkehard Plesser, UMB (hans.ekkehard.plesser at umb.no)
"""

preamble = r"""
\documentclass{article}
\usepackage[margin=1in]{geometry}
\usepackage{tabularx}
\usepackage{multirow}
\usepackage{colortbl}

\usepackage[fleqn]{amsmath}
\setlength{\mathindent}{0em}
%%\usepackage{mathpazo}
\usepackage{breqn}

\usepackage[scaled=.95]{helvet}
\renewcommand\familydefault{\sfdefault}

\renewcommand\arraystretch{1.2}
\pagestyle{empty}

\newcommand{\hdr}[3]{
    \multicolumn{#1}{|l|}{
        \color{white}\cellcolor[gray]{0.0}
        \textbf{\makebox[0pt]{#2}\hspace{0.5\linewidth}\makebox[0pt][c]{#3}}
    }
}

\begin{document}
"""

summary_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|l|X|}\hline
\hdr{2}{A}{Model Summary}\\ \hline
\textbf{Populations}     & %(population_names)s \\ \hline
\textbf{Topology}        & --- \\ \hline
\textbf{Connectivity}    & %(connectivity)s \\ \hline
\textbf{Neuron models}   & %(neuron_models)s \\ \hline
\textbf{Channel models}  & --- \\ \hline
\textbf{Synapse models}  & --- \\ \hline
\textbf{Plasticity}      & %(synapse_models)s\\ \hline
\textbf{Input}           & %(input_types)s \\ \hline
\textbf{Measurements}    & %(measurements)s \\ \hline
\end{tabularx}

\vspace{2ex}
"""

populations_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|l|l|X|}\hline
\hdr{3}{B}{Populations}\\ \hline
    \textbf{Name}   & \textbf{Elements} & \textbf{Size} \\ \hline
%(populations_description)s
\end{tabularx}

\vspace{2ex}
"""

connectivity_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|l|l|l|X|X|}\hline
\hdr{5}{C}{Connectivity}\\ \hline
\textbf{Source} & \textbf{Destination} & \textbf{Target} & \textbf{Synapse} & \textbf{Pattern} \\ \hline
%(projections_description)s
\end{tabularx}

\vspace{2ex}
"""


parameters_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|X|}\hline
\hdr{1}{F}{Parameters}\\ \hline
\\ \hline
\end{tabularx}
\vspace{2ex}
"""

constants_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.25\linewidth}|p{0.25\linewidth}|X|}\hline
\textbf{Constants} &\textbf{Name} & \textbf{Value}   \\ \hline
%(parameters)s
\end{tabularx}

\vspace{2ex}
"""

functions_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.25\linewidth}|X|}\hline
\textbf{Functions} &  
%(parameters)s
\\ \hline
\end{tabularx}

\vspace{2ex}
"""

popparameters_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.25\linewidth}|p{0.25\linewidth}|X|}\hline
\textbf{Population} & \textbf{Parameter} & \textbf{Value}   \\ \hline
%(parameters)s
\end{tabularx}

\vspace{2ex}
"""

projparameters_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.25\linewidth}|p{0.25\linewidth}|X|}\hline
\textbf{Projection} & \textbf{Parameter} & \textbf{Value}   \\ \hline
%(parameters)s
\end{tabularx}

\vspace{2ex}
"""

input_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|l|X|}\hline
\hdr{2}{G}{Input}\\ \hline
\textbf{Name} & \textbf{Description} \\ \hline

"""

footer = r"""
\end{document}
"""

##################################
### Main method
##################################

def report_latex(filename="./report.tex", standalone=True, gather_subprojections=False, net_id=0):
    """   
    Generates a .tex file describing the network according to:

    > Nordlie E, Gewaltig M-O, Plesser HE (2009). Towards Reproducible Descriptions of Neuronal Network Models. PLoS Comput Biol 5(8): e1000456.

    :param filename: name of the .tex file where the report will be written (default: "./report.tex")
    :param standalone: tells if the generated file should be directly compilable or only includable (default: True)
    :param gather_subprojections: if a projection between two populations has been implemented as a multiple of projections between sub-populations, this flag allows to group them in the summary (default: False).
    """

    # stdout
    Messages._print('Generating report in', filename)

    # Generate the summary
    summary = _generate_summary(net_id)
    # Generate the populations
    populations = _generate_populations(net_id)
    # Generate the projections
    projections = _generate_projections(net_id, gather_subprojections)
    # Generate the neuron models
    neuron_models = _generate_neuron_models(net_id)
    # Generate the synapse models
    synapse_models = _generate_synapse_models(net_id)
    # Generate the constants
    constants = _generate_constants(net_id)
    # Generate the functions
    functions = _generate_functions(net_id)
    # Generate the population parameters
    pop_parameters = _generate_population_parameters(net_id)
    # Generate the population parameters
    proj_parameters = _generate_projection_parameters(net_id, gather_subprojections)
    # Generate a list of input layers (Poisson generators, SpikeSourceArray, TimedArray, etc)
    inputs = _generate_inputs(net_id)
    # Generate the measurements (in our case the BOLD recording)
    measurements = _generate_measurements(net_id)

    # Possibly create the directory if it does not exist
    path_name = os.path.dirname(filename)
    if not path_name in ["", "."]:
        if not os.path.exists(path_name):
            os.makedirs(path_name)

    with open(filename, 'w') as wfile:
        if standalone:
            wfile.write(header)
            wfile.write(preamble)
        wfile.write(summary)
        wfile.write(populations)
        wfile.write(projections)
        wfile.write(neuron_models)
        wfile.write(synapse_models)
        wfile.write(parameters_template)
        wfile.write(constants)
        wfile.write(functions)
        wfile.write(pop_parameters)
        wfile.write(proj_parameters)
        wfile.write(inputs)
        wfile.write(measurements)
        if standalone:
            wfile.write(footer)

##################################
### Process major fields
##################################

def _generate_summary(net_id):
    "part A"

    population_names = str(len(NetworkManager().get_network(net_id=net_id).get_populations())) + ': '
    connectivity = ""
    neuron_models = ""
    synapse_models = ""

    # List the names of all populations
    for pop in NetworkManager().get_network(net_id=net_id).get_populations():
        # population name
        population_names += LatexParser.pop_name(pop.name) + ", "
    population_names = population_names[:-2] # suppress the last ,

    # List all BOLD recordings
    measurements = ""

    # List all neuron types
    neuron_model_names = []
    input_type_list = []

    for neur in GlobalObjectManager().get_neuron_types():
        # bold models sorted in measurements
        if isinstance(neur, BoldModel):
            if neur._model_instantiated:
                measurements += neur.name + ', '
        # specific inputs sorted in inputs
        if neur.name in input_name_list:
            input_type_list.append(neur.name)
        # neuron model
        else:
            neuron_model_names.append(neur.name)
    for neur in list(set(neuron_model_names)):
        neuron_models += neur + ', '
    # suppress the last ,
    measurements = measurements[:-2]
    neuron_models = neuron_models[:-2]
    # escape "_" in names
    neuron_models = neuron_models.replace("_", r"\_")

    # Gather input types
    if len(input_type_list) > 0:
        input_types = ""
        for input in list(set(input_type_list)):
            input_types += input + ', '
        input_types = input_types[:-2]
    else:
        input_types = "---"

    list_connectivity = []
    list_synapse_models = []
    for proj in NetworkManager().get_network(net_id=net_id).get_projections():
        list_connectivity.append(proj.connector_name)
        if not proj.synapse_type.name in list(Synapse._default_names.values()) + ['-']:
            list_synapse_models.append(proj.synapse_type.name)
    for con in list(set(list_connectivity)):
        connectivity += con + ', '
    for syn in list(set(list_synapse_models)):
        synapse_models += syn + ', '
    connectivity = connectivity[:-2]
    synapse_models = synapse_models[:-2] # suppress the last ,

    # Write the summary
    txt = summary_template  % {
        'population_names' : population_names,
        'connectivity' : connectivity,
        'neuron_models' : neuron_models,
        'synapse_models' : synapse_models,
        'input_types': input_types,
        'measurements': measurements
    }
    return txt

def _generate_populations(net_id):
    def format_size(pop):
        size = str(pop.size)
        if pop.dimension >1:
            size += ' ('
            for d in range(pop.dimension):
                size += str(pop.geometry[d]) + '*'
            size = size.rsplit('*', 1)[0] + ')'
        return size

    txt = ""
    pop_tpl = r"""
    %(pop_name)s             & %(neuron_type)s        & $N_{\text{%(pop_name)s}}$ = %(size)s  \\ \hline
"""
    for pop in NetworkManager().get_network(net_id=net_id).get_populations():
        # Find a name for the neuron
        if pop.neuron_type.name in Neuron._default_names.values(): # name not set
            neuron_name = "Neuron " + str(pop.neuron_type._rk_neurons_type)
        else:
            neuron_name = pop.neuron_type.name
            neuron_name = neuron_name.replace("_", r"\_")    # escape "_" in names

        txt += pop_tpl % {
            'pop_name': LatexParser.pop_name(pop.name), 
            'neuron_type': neuron_name, 
            'size': format_size(pop)}

    return populations_template % {'populations_description': txt}

def _generate_constants(net_id):
    cst_tpl = r"""
    & $%(param)s$        & %(value)s  \\\\ \\hline
"""
    parameters = ""

    constants = NetworkManager().get_network(net_id).get_constants()

    if len(constants) == 0:
        return ""

    for constant in constants:
        parameters += cst_tpl % {'param': LatexParser._latexify_name(constant.name, []), 'value': constant.value}

    txt = constants_template % {'parameters': parameters}

    return txt


def _generate_functions(net_id):

    functions = ""
    if GlobalObjectManager().number_functions() == 0:
        return functions

    for _, func in GlobalObjectManager().get_functions():
        functions += LatexParser._process_functions(func, net_id) + "\n"

    return functions_template % {
        'parameters': functions, 
        'firstfunction': r"\hdr{1}{G}{Functions}\\ \hline"}

def _generate_population_parameters(net_id):
    txt = ""
    pop_tpl = r"""
    %(name)s             & $%(param)s$        & %(value)s  \\ \hline
"""
    for rk, pop in enumerate(NetworkManager().get_network(net_id=net_id).get_populations()):
        parameters = ""
        for idx, param in enumerate(pop.parameters):
            val = pop.init[param]
            if isinstance(val, (list, np.ndarray)):
                val = "$[" + str(np.array(val).min()) + ", " + str(np.array(val).max()) + "]$"
            parameters += pop_tpl % {'name': LatexParser.pop_name(pop.name) if idx==0 else "", 'param': LatexParser._latexify_name(param, []), 'value': val}

        if parameters != "":
            txt += popparameters_template % {'parameters': parameters, 'firstpopulation': r"\hdr{3}{H}{Population parameters}\\ \hline" if rk==0 else ""}

    return txt

def _generate_projections(net_id, gather_subprojections):
    txt = r""
    proj_tpl = r"""
    %(pre)s & %(post)s & %(target)s & %(synapse)s &
    %(description)s \\ \hline
"""
    if gather_subprojections:
        projections = []
        for proj in NetworkManager().get_network(net_id=net_id).get_projections():
            for existing_proj in projections:
                if proj.pre.name == existing_proj.pre.name and proj.post.name == existing_proj.post.name and proj.target == existing_proj.target : # TODO
                    break
            else:
                projections.append(proj)
    else:
        projections = NetworkManager().get_network(net_id=net_id).get_projections()

    for proj in projections:
        # Find a name for the synapse
        if proj.synapse_type.name in Synapse._default_names.values(): # name not set
            synapse_name = "Synapse " + str(proj.synapse_type._rk_synapses_type)
        else:
            synapse_name = proj.synapse_type.name

        txt += proj_tpl % { 'pre': LatexParser.pop_name(proj.pre.name), 
                            'post': LatexParser.pop_name(proj.post.name), 
                            'target': (LatexParser._format_list(proj.target, ' / ')).replace("_", r"\_"),
                            'synapse': synapse_name,
                            'description': proj.connector_description.replace("_", r"\_")}

    return connectivity_template % {'projections_description': txt}

def _generate_projection_parameters(net_id, gather_subprojections):
    txt = ""
    proj_tpl = r"""
    %(name)s & $%(param)s$        & %(value)s  \\ \hline
"""
    if gather_subprojections:
        projections = []
        for proj in NetworkManager().get_network(net_id=net_id).get_projections():
            for existing_proj in projections:
                if proj.pre.name == existing_proj.pre.name and proj.post.name == existing_proj.post.name and proj.target == existing_proj.target : # TODO
                    break
            else:
                projections.append(proj)
    else:
        projections = NetworkManager().get_network(net_id=net_id).get_projections()

    first = True
    for rk, proj in enumerate(projections):
        parameters = ""
        for idx, param in enumerate(proj.parameters):
            if param == 'w':
                continue
            if idx == 0:
                proj_name = "%(pre)s  $\\rightarrow$ %(post)s with target %(target)s" % {
                    'pre': LatexParser.pop_name(proj.pre.name), 
                    'post': LatexParser.pop_name(proj.post.name), 
                    'target': LatexParser._format_list(proj.target, ' / ')}
            else:
                proj_name = ""
            val = proj.init[param]
            
            if isinstance(val, (list, np.ndarray)):
                val = "$[" + str(np.min(val)) + ", " + str(np.max(val)) + "]$"
            parameters += proj_tpl % {'name': proj_name, 'param': LatexParser._latexify_name(param, []), 'value': val}

        if parameters != "":
            txt += projparameters_template % {
                'parameters': parameters, 
                'firstprojection': r"\hdr{3}{H}{Projection parameters}\\ \hline" if first else ""}
            first = False

    return txt

def _generate_neuron_models(net_id):
    neurons = ""

    firstneuron = r"\hdr{2}{D}{Neuron Models}\\ \hline"

    neuron_tpl = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.15\linewidth}|X|}\hline
%(firstneuron)s
\textbf{Name} & %(name)s \\ \hline
\textbf{Type} & %(description)s\\ \hline
\textbf{%(equation_type)s} &
%(variables)s
%(spike)s
%(functions)s
\end{tabularx}
\vspace{2ex}
"""

    first_neuron = True
    for neuron in GlobalObjectManager().get_neuron_types():

        # skip bold models
        if isinstance(neuron, BoldModel):
            continue

        # Some input layers have no meaningful model descriptions
        if neuron.name in ["Timed Array", "Input Array"]:
            continue

        # Name
        if neuron.name in Neuron._default_names.values(): # name not set
            neuron_name = "Neuron " + str(neuron._rk_neurons_type)
        else:
            neuron_name = neuron.name
            neuron_name = neuron_name.replace("_", r"\_")  # escape "_" in names

        # Generate the code for the equations
        variables, spike_condition, spike_reset = LatexParser._process_neuron_equations(neuron, net_id=net_id)

        eqs = ""
        for var in variables:
            eqs += r"""
\begin{dmath*}
%(eq)s
\end{dmath*}
""" % {'eq': var['latex']}

        variables_eqs = r"""
%(eqs)s
\\ \hline
""" % {'eqs': eqs}

        # Spiking neurons have an extra field for the spike condition
        spike_extra = ""
        if neuron.type == 'spike':
            spike_code = "If $" + spike_condition + r"$ or $t \leq t^* + t_{\text{refractory}}$:"

            # Reset
            spike_code += r"""
            \begin{enumerate}
                \item Emit a spike at time $t^*$"""

            for var in spike_reset:
                spike_code += r"""
            \item $""" + var + "$"

            spike_code += r"""
        \end{enumerate}"""


            spike_extra = r"""
\textbf{Spiking} &
%(spike)s
\\ \hline
""" % {'spike': spike_code}

        # Possible function
        functions = ""
        if not neuron.functions == None:
            functions = r"""
\textbf{Functions} &
%(functions)s
\\ \hline
""" % {'functions': LatexParser._process_functions(neuron.functions)}

        # Build the dictionary
        desc = {
            'name': neuron_name,
            'description': neuron.short_description,
            'firstneuron': firstneuron if first_neuron else "",
            'variables': variables_eqs,
            'spike': spike_extra,
            'functions': functions,
            'equation_type': "Subthreshold dynamics" if neuron.type == 'spike' else 'Equations'
        }

        # Generate the code depending on the neuron position
        neurons += neuron_tpl % desc

        # first_neuron = False, to prevent multiple header
        first_neuron = False

    return neurons

def _generate_synapse_models(net_id):
    firstsynapse = ""
    synapses = ""

    firstsynapse = r"\hdr{2}{E}{Synapse Models}\\ \hline"

    synapse_tpl = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.15\linewidth}|X|}\hline
%(firstsynapse)s
\textbf{Name} & %(name)s \\ \hline
\textbf{Type} & %(description)s\\ \hline
%(psp)s
%(variables)s
%(preevent)s
%(postevent)s
%(functions)s
\end{tabularx}
\vspace{2ex}
"""
    for idx, synapse in enumerate(GlobalObjectManager().get_synapse_types()):
        # Do not document default synapses
        if synapse.name == "-":
            continue

        # Find a name for the synapse
        if synapse.name in Synapse._default_names.values(): # name not set
            synapse_name = "Synapse " + str(synapse._rk_synapses_type)
        else:
            synapse_name = synapse.name

        # Generate the code for the equations
        psp, variables, pre_desc, post_desc = LatexParser._process_synapse_equations(synapse, net_id)

        eqs = ""
        for var in variables:
            eqs += r"""
\begin{dmath*}
%(eq)s
\end{dmath*}
""" % {'eq': var['latex']}

        variables_eqs = r"""
%(eqs)s
\\ \hline
""" % {'eqs': eqs}

        # Synaptic variables
        variables = r"""
\textbf{Equations} & %(variables)s  
\\ \hline""" % {'variables':eqs} if eqs != "" else ""

        # PSP
        if psp != "":
            psp_code = r"""
\textbf{PSP} & \begin{dmath*}
%(psp)s
\end{dmath*}
\\ \hline""" % {'psp': psp}
        else:
            psp_code = ""

        # Spiking neurons have extra fields for the event-driven
        if synapse.type == 'spike':
            if len(pre_desc) > 0:
                txt_pre = ""
                for l in pre_desc:
                    txt_pre += r"""
\begin{dmath*}
%(eq)s
\end{dmath*}
""" % {'eq': l}
                preevent = r"""
\textbf{Pre-synaptic event} &
%(preevent)s
\\ \hline
""" % {'preevent': txt_pre}
            else:
                preevent = ""

            if len(post_desc) > 0:
                txt_post = ""
                for l in post_desc:
                    txt_post += r"""
\begin{dmath*}
%(eq)s
\end{dmath*}
""" % {'eq': l}
                postevent = r"""
\textbf{Post-synaptic event} &
%(postevent)s
\\ \hline
""" % {'postevent': txt_post}
            else:
                postevent = ""
        else:
            preevent = ""
            postevent = ""

        # Possible functions
        functions = ""
        if not synapse.functions == None:
            functions = r"""
\textbf{Functions} &
%(functions)s
\\ \hline
""" % {'functions': LatexParser._process_functions(synapse.functions)}

        # Build the dictionary
        desc = {
            'name': synapse_name,
            'description': synapse.short_description,
            'firstsynapse': firstsynapse if idx == 0 else "",
            'variables': variables,
            'psp': psp_code,
            'preevent': preevent,
            'postevent': postevent,
            'functions': functions
        }

        # Generate the code
        synapses += synapse_tpl % desc

    return synapses

def _generate_inputs(net_id):
    input_desc = input_template

    for pop in NetworkManager().get_network(net_id=net_id).get_populations():

        if isinstance(pop, input_type_list):
            input_desc += r"{} & {} \\ \hline".format(pop.name, pop.neuron_type.short_description)
            input_desc += '\n'

    input_desc += r"""
\end{tabularx}

\vspace{2ex}
"""
    return input_desc

def _generate_measurements(net_id):
    measurements = ""

    header = r"\hdr{2}{D}{Bold Measurement Models}\\ \hline"

    measurements_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.15\linewidth}|X|}\hline
%(header)s
\textbf{Name} & %(name)s \\ \hline
\textbf{Type} & %(description)s\\ \hline
\textbf{%(equation_type)s} &
%(variables)s
\end{tabularx}
\vspace{2ex}
"""

    first_define = True
    for neuron in GlobalObjectManager().get_neuron_types():

        # skip non-bold models
        if not isinstance(neuron, BoldModel):
            continue

        # the model is not used
        if not neuron._model_instantiated:
            continue

        # Name
        if neuron.name in Neuron._default_names.values(): # name not set
            neuron_name = "Neuron " + str(neuron._rk_neurons_type)
        else:
            neuron_name = neuron.name

        # Generate the code for the equations
        variables, _, _ = LatexParser._process_neuron_equations(neuron, net_id=net_id)

        eqs = ""
        for var in variables:
            eqs += r"""
\begin{dmath*}
%(eq)s
\end{dmath*}
""" % {'eq': var['latex']}

        variables_eqs = r"""
%(eqs)s
\\ \hline
""" % {'eqs': eqs}

        # Build the dictionary
        desc = {
            'name': neuron_name,
            'description': neuron.short_description,
            'header': header if first_define else "",
            'variables': variables_eqs,
            'equation_type': "Equations"
        }

        # Generate the code depending on the neuron position
        measurements += measurements_template % desc

        # first_define disable to prevent multiple table header
        first_define = False

    # no models were processed
    if len(measurements) == 0:
        measurements = r"""
\noindent
\begin{tabularx}{\linewidth}{|l|X|}\hline
\hdr{2}{H}{Measurement}\\ \hline
\textbf{Type} & \textbf{Description} \\ \hline
---
\\ \hline
\end{tabularx}
"""


    return measurements



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import ANNarchy

# ANNarchy objects
from ANNarchy.core import Global
from ANNarchy.core.PopulationView import PopulationView
from ANNarchy.models.Synapses import DefaultRateCodedSynapse
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_precision
from ANNarchy.intern import Messages

# Code templates
from ANNarchy.generator.Projection.ProjectionGenerator import ProjectionGenerator, get_bounds
from ANNarchy.generator.Projection.OpenMP import *

# Useful functions
from ANNarchy.generator.Utils import generate_equation_code, tabify, remove_trailing_spaces, get_highest_available_simd_instructions_type, determine_idx_type_for_projection

import re
from copy import deepcopy

class OpenMPGenerator(ProjectionGenerator):
    """
    Generate the header for a Population object to run either on single core
    or multi-cores with OpenMP.
    """
    def __init__(self, profile_generator, net_id):
        # The super here calls all the base classes, so first
        # ProjectionGenerator and afterwards OpenMPConnectivity
        # TODO: this is python 2 syntax
        super(OpenMPGenerator, self).__init__(profile_generator, net_id)

    def header_struct(self, proj, annarchy_dir):
        """
        Generate the projection header for a given projection. The resulting
        code will be stored in a file called proj<unique_id>.hpp in the
        directory indicated by annarchy_dir.

        This function will be called from CodeGenerator.

        Returns:

        * proj_desc: a dictionary with all call statements for the distinct
        operations (like compute_psp, update_synapse, etc.)
        """
        # Initial state
        self._templates = deepcopy(BaseTemplates.openmp_templates)
        self._template_ids = {}

        # Select the C++ connectivity template
        sparse_matrix_include, sparse_matrix_format, sparse_matrix_args, single_matrix = self._select_sparse_matrix_format(proj, suppress_printouts=False)

        # Update template fill elements
        self._configure_template_ids(proj, single_matrix)

        # Generate declarations and accessors for the variables
        decl, accessor = self._declaration_accessors(proj, single_matrix)

        # Initialize the projection
        init_weights, init_delays, init_parameters_variables = self._init_parameters_variables(proj, single_matrix)

        # Synaptic plasticity
        update_prefix, update_variables = self._update_synapse(proj, single_matrix)

        # Update the random distributions
        init_rng = self._init_random_distributions(proj)
        update_rng = self._update_random_distributions(proj)

        post_event_prefix, post_event = self._post_event(proj, single_matrix)

        # Compute sum is the trickiest part
        if proj.synapse_type.type == 'rate':
            psp_prefix, psp_code = self._computesum_rate(proj, single_matrix)
        else:
            psp_prefix, psp_code = self._computesum_spiking(proj, single_matrix)

        # Detect event-driven variables
        has_event_driven = False
        for var in proj.synapse_type.description['variables']:
            if var['method'] == 'event-driven':
                has_event_driven = True

        # Detect delays to generate the code
        has_delay = proj.max_delay > 1
        if has_delay:
            update_max_delay, reset_ring_buffer = self._update_max_delay(proj)
        else:
            update_max_delay = ""
            reset_ring_buffer = ""

        # Some Connectivity implementations requires the number of threads in constructor
        if ConfigManager().get('num_threads', self._net_id) > 1:
            if proj._storage_format == "lil":
                if single_matrix or proj._no_split_matrix:
                    num_threads_acc = ""
                else:
                    num_threads_acc = ", global_num_threads"
            elif proj._storage_format == "csr":
                if proj._no_split_matrix or single_matrix:
                    num_threads_acc = ""
                else:
                    num_threads_acc = ", global_num_threads"
            else:
                num_threads_acc = ""
        else:
            num_threads_acc = ""

        # Connectivity template
        if 'declare_connectivity_matrix' not in proj._specific_template.keys():
            connector_call = self._connectivity_init(proj, sparse_matrix_format, sparse_matrix_args) % {
                'sparse_format': sparse_matrix_format,
                'init_weights': init_weights,
                'init_delays': init_delays,
                'rng_idx': "[0]" if single_matrix else "",
                'add_args': "",
                'num_threads': num_threads_acc,
                'float_prec': ConfigManager().get('precision', self._net_id),
                'idx_type': self._template_ids['idx_type']
            }
            declare_connectivity_matrix = ""
            access_connectivity_matrix = ""
        else:
            sparse_matrix_format = "SpecificConnectivity"
            sparse_matrix_args = ""
            sparse_matrix_include = "#include \"Specific.hpp\"\n"
            connector_call = ""
            declare_connectivity_matrix = proj._specific_template['declare_connectivity_matrix']
            access_connectivity_matrix = proj._specific_template['access_connectivity_matrix']

        # local functions
        decl['parameters_variables'] += self._local_functions(proj)

        # Memory management
        size_in_bytes = self._size_in_bytes(proj)
        clear_container = self._clear_container(proj)

        # Structural plasiticity
        creating = self.creating(proj)
        pruning = self.pruning(proj)

        # Profiling
        if self._prof_gen:
            include_profile = """#include "Profiling.hpp"\n"""
            declare_profile, init_profile = self._prof_gen.generate_init_projection(proj)
        else:
            include_profile = ""
            init_profile = ""
            declare_profile = ""

        # Sparse matrix format specific
        # HD (20th May 2022): this is probably not the best way to do it ...
        declare_additional = decl['additional']
        init_additional = ""
        if proj._storage_format == "dense" and proj._storage_order == "pre_to_post":
            declare_additional += """\t// dense matrix - static schedule
    std::vector<int> post_slices_;
        """
            init_additional += """\t// static distribution across threads
        int chunk_size = static_cast<int>(ceil(static_cast<double>(this->num_rows_) / static_cast<double>(global_num_threads)));
        post_slices_ = std::vector<int>(1, 0);
        for (int t = 1; t <= global_num_threads; t++)
            post_slices_.push_back(std::min<int>(t*chunk_size, this->num_rows_));
"""

        # Additional info (overwritten)
        include_additional = ""
        struct_additional = ""
        access_additional = ""
        if 'include_additional' in proj._specific_template.keys():
            include_additional = proj._specific_template['include_additional']
        if 'struct_additional' in proj._specific_template.keys():
            struct_additional = proj._specific_template['struct_additional']
        if 'init_additional' in proj._specific_template.keys():
            init_additional = proj._specific_template['init_additional']
        if 'access_additional' in proj._specific_template.keys():
            access_additional = proj._specific_template['access_additional']

        final_code_dict = {
            # version tag
            'annarchy_version': ANNarchy.__release__,
            # fill code templates
            'id_pre': proj.pre.id,
            'id_post': proj.post.id,
            'id_proj': proj.id,
            'name_pre': proj.pre.name,
            'name_post': proj.post.name,
            'target': proj.target,
            'sparse_matrix_include': sparse_matrix_include,
            'include_additional': include_additional,
            'include_profile': include_profile,
            'struct_additional': struct_additional,
            'declare_connectivity_matrix': declare_connectivity_matrix,
            'access_connectivity_matrix': access_connectivity_matrix,
            'declare_delays': decl['declare_delay'] if has_delay else "",
            'declare_event_driven': decl['event_driven'] if has_event_driven else "",
            'declare_rng': decl['rng'],
            'declare_parameters_variables': decl['parameters_variables'],
            'declare_additional': declare_additional,
            'declare_profile': declare_profile,
            'connector_call': connector_call,
            'init_event_driven': "",
            'init_rng': init_rng,
            'init_parameters_variables': init_parameters_variables,
            'init_additional': init_additional,
            'init_profile': init_profile,
            'psp_prefix': psp_prefix,
            'psp_code': psp_code,
            'update_rng': update_rng,
            'update_prefix': update_prefix,
            'update_variables': update_variables,
            'update_max_delay': update_max_delay,
            'reset_ring_buffer': reset_ring_buffer,
            'post_event_prefix': post_event_prefix,
            'post_event': post_event,
            'access_parameters_variables': accessor,
            'access_additional': access_additional,
            'size_in_bytes': size_in_bytes,
            'clear_container': clear_container,
            'sparse_format': sparse_matrix_format,
            'sparse_format_args': sparse_matrix_args,
            'float_prec': ConfigManager().get('precision', self._net_id),
            'creating': creating,
            'pruning': pruning
        }

        final_code = self._templates['projection_header'] % final_code_dict

        # remove right-trailing white spaces
        final_code = remove_trailing_spaces(final_code)

        # Store file
        with open(annarchy_dir+'/generate/net'+str(self._net_id)+'/proj'+str(proj.id)+'.hpp', 'w') as ofile:
            ofile.write(final_code)

        # Dictionary for inclusions in ANNarchy.cpp
        proj_desc = {
            'include': """#include "proj%(id)s.hpp"\n""" % {'id': proj.id},
            'extern': """extern ProjStruct%(id)s *proj%(id)s;\n"""% {'id': proj.id},
            'instance': """ProjStruct%(id)s *proj%(id)s;\n"""% {'id': proj.id},
            'init': """    proj%(id)s->init_projection();\n""" % {'id' : proj.id}
        }

        proj_desc['compute_psp'] = """\tproj%(id)s->compute_psp(tid, nt);\n""" % {'id' : proj.id}
        proj_desc['update'] = "" if update_variables == "" else """\tproj%(id)s->update_synapse(tid, nt);\n""" % {'id': proj.id}
        proj_desc['rng_update'] = "" if update_rng == "" else """\tproj%(id)s->update_rng();\n""" % {'id': proj.id}
        proj_desc['post_event'] = "" if post_event == "" else """\tproj%(id)s->post_event(tid, nt);\n""" % {'id': proj.id}

        return proj_desc

    def _configure_template_ids(self, proj, single_matrix):
        """
        Assign the correct template dictionary based on projection
        storage format. Also sets the basic template ids, e. g. indices
        """
        # Index data types depend on the matrix dimension
        # HD (1st Dec. 2021):   until now, the data type optimization is disabled for
        #                       spiking models. I want to adjust new codes already as I
        #                       hope to update the spike code generation soon ...
        idx_type, _, size_type, _ = determine_idx_type_for_projection(proj)

        self._template_ids.update({
            'id_proj' : proj.id,
            'target': proj.target,
            'id_post': proj.post.id,
            'id_pre': proj.pre.id,
            'float_prec': ConfigManager().get('precision', self._net_id),
            'pre_prefix': 'pop'+ str(proj.pre.id) + '->',
            'post_prefix': 'pop'+ str(proj.post.id) + '->',
            'idx_type': idx_type,
            'size_type': size_type
        })

        if proj._storage_format == "lil":
            if proj.synapse_type.type == "rate":
                # Rate-coded models LIL
                if single_matrix:
                    self._templates.update(LIL_OpenMP.conn_templates)
                    self._template_ids.update(LIL_OpenMP.conn_ids)
                else:
                    self._templates.update(LIL_Sliced_OpenMP.conn_templates)
                    self._template_ids.update(LIL_Sliced_OpenMP.conn_ids)

            else:
                # Spiking models LIL
                if proj._storage_order == "pre_to_post":
                    raise NotImplementedError

                if single_matrix:
                    self._templates.update(LIL_OpenMP.conn_templates)
                    self._template_ids.update(LIL_OpenMP.conn_ids)
                else:
                    self._templates.update(LIL_Sliced_OpenMP.conn_templates)
                    self._template_ids.update(LIL_Sliced_OpenMP.conn_ids)

        elif proj._storage_format == "bsr":
            if proj.synapse_type.type == "rate":
                if single_matrix:
                    self._templates.update(BSR_OpenMP.conn_templates)
                    self._template_ids.update(BSR_OpenMP.conn_ids)
                else:
                    raise NotImplementedError

            else:
                raise NotImplementedError

        elif proj._storage_format == "csr":
            if proj.synapse_type.type == "rate":
                self._templates.update(CSR_OpenMP.conn_templates)
                self._template_ids.update(CSR_OpenMP.conn_ids)

            else:
                # Spiking models (CSRC)
                if proj._storage_order == "post_to_pre":
                    if single_matrix:
                        self._templates.update(CSR_OpenMP.conn_templates)
                        self._template_ids.update(CSR_OpenMP.conn_ids)
                    else:
                        self._templates.update(CSR_Sliced_OpenMP.conn_templates)
                        self._template_ids.update(CSR_Sliced_OpenMP.conn_ids)

                else:
                    if single_matrix:
                        self._templates.update(CSR_T_OpenMP.conn_templates)
                        self._template_ids.update(CSR_T_OpenMP.conn_ids)
                    else:
                        self._templates.update(CSR_T_Sliced_OpenMP.conn_templates)
                        self._template_ids.update(CSR_T_Sliced_OpenMP.conn_ids)

        elif proj._storage_format == "coo":
            if proj.synapse_type.type == "rate":
                # Rate-coded models coordinate format
                if single_matrix:
                    self._templates.update(COO_OpenMP.conn_templates)
                    self._template_ids.update(COO_OpenMP.conn_ids)
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError

        elif proj._storage_format == "dia":
            if proj.synapse_type.type == "rate":
                # Rate-coded models coordinate format
                if single_matrix:
                    self._templates.update(DIA_OpenMP.conn_templates)
                    self._template_ids.update(DIA_OpenMP.conn_ids)
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError

        elif proj._storage_format == "ellr":
            if proj.synapse_type.type == "rate":
                # Rate-coded models ELLPACK-R format
                if single_matrix:
                    self._templates.update(ELLR_OpenMP.conn_templates)
                    self._template_ids.update(ELLR_OpenMP.conn_ids)
                else:
                    raise NotImplementedError

            else:
                raise Global.InvalidConfiguration("    "+proj.name+": ELLPACK-R format is not available for spiking models.")

        elif proj._storage_format == "sell":
            if proj.synapse_type.type == "rate":
                # Rate-coded models sliced ELLPACK-R format
                if single_matrix:
                    self._templates.update(SELL_OpenMP.conn_templates)
                    self._template_ids.update(SELL_OpenMP.conn_ids)
                else:
                    raise NotImplementedError

            else:
                raise Global.InvalidConfiguration("    "+proj.name+": sliced ELLPACK format is not available for spiking models.")

        elif proj._storage_format == "ell":
            if proj.synapse_type.type == "rate":
                # Rate-coded models ELLPACK format
                if single_matrix:
                    self._templates.update(ELL_OpenMP.conn_templates)
                    self._template_ids.update(ELL_OpenMP.conn_ids)
                else:
                    raise NotImplementedError

            else:
                raise Global.InvalidConfiguration("    "+proj.name+": ELLPACK format is not available for spiking models.")

        elif proj._storage_format == "dense":
            if proj._storage_order == "post_to_pre":
                self._templates.update(Dense_OpenMP.conn_templates)
                self._template_ids.update(Dense_OpenMP.conn_ids)
            else:
                self._templates.update(Dense_T_OpenMP.conn_templates)
                self._template_ids.update(Dense_T_OpenMP.conn_ids)

        else:
            raise NotImplementedError

    def creating(self, proj):
        """
        Synapse creation based on neural state variables.

        TODO: documentation
        """
        if 'creating' not in proj.synapse_type.description.keys():
            return ""

        if proj._storage_format != "lil":
            raise NotImplementedError("Structural plasticity is only available for LIL structures.")

        creating_structure = proj.synapse_type.description['creating']

        # Random stuff
        proba = ""
        proba_init = ""
        if 'proba' in creating_structure['bounds'].keys():
            val = creating_structure['bounds']['proba']
            proba += '&&(unif(rng[0])<' + val + ')'
            proba_init += "std::uniform_real_distribution<double> unif(0.0, 1.0);"
        if  creating_structure['rd']:
            proba_init += "\n        " +  creating_structure['rd']['template'] + ' rd(' + creating_structure['rd']['args'] + ');'

        # delays
        delay = ""
        if 'd' in creating_structure['bounds'].keys():
            d = int(creating_structure['bounds']['delay']/ConfigManager().get('dt', self._net_id))
            if proj.max_delay > 1 and proj.uniform_delay == -1:
                if d > proj.max_delay:
                    Messages._error('creating: you can not add a delay higher than the maximum of existing delays')

                delay = ", " + str(d)
            else:
                if d != proj.uniform_delay:
                    Messages._error('creating: you can not add a delay different from the others if they were constant.')

        creation_ids = deepcopy(self._template_ids)
        creation_ids.update({
            'pre_index': '[rk_pre]',
            'post_index': '[rk_post]'
        })
        creating_condition = creating_structure['cpp'] % creation_ids
        creation_ids.update({
            'eq': creating_structure['eq'],
            'condition': creating_condition,
            'weights': 0.0 if not 'w' in creating_structure['bounds'].keys() else creating_structure['bounds']['w'],
            'proba' : proba, 'proba_init': proba_init,
            'delay': delay
        })
        creating = self._templates['structural_plasticity']['create'] % creation_ids

        return creating

    def pruning(self, proj):
        """
        Synapse removal based on neural state variables.

        TODO: documentation
        """
        if 'pruning' not in proj.synapse_type.description.keys():
            return ""

        if proj._storage_format != "lil":
            raise NotImplementedError("Structural plasticity is only available for LIL structures.")

        pruning_structure = proj.synapse_type.description['pruning']

        proba = ""
        proba_init = ""
        if 'proba' in pruning_structure['bounds'].keys():
            val = pruning_structure['bounds']['proba']
            proba = '&&(unif(rng[0])<' + val + ')'
            proba_init = "std::uniform_real_distribution<double> unif(0.0, 1.0);"
        if pruning_structure['rd']:
            proba_init += "\n        " +  pruning_structure['rd']['template'] + ' rd(' + pruning_structure['rd']['args'] + ');'

        pruning_ids = deepcopy(self._template_ids)
        pruning_condition = pruning_structure['cpp'] % pruning_ids

        pruning_ids.update({
            'eq': pruning_structure['eq'],
            'condition': pruning_condition,
            'proba' : proba,
            'proba_init': proba_init
        })
        pruning = self._templates['structural_plasticity']['prune'] % pruning_ids

        return pruning

    def _computesum_rate(self, proj, single_matrix):
        """
        Create the c++ code for post-synaptic potential computation. 

        Parameters:

        * proj:             Instance of ANNarchy.core.Projection
        * single_matrix:    Is the basic connectivity a complete matrix or sliced
        """
        # Specific projection (e. g. convolution, pooling or user-defined stuff)
        if 'psp_code' in proj._specific_template.keys():
            psp_code = proj._specific_template['psp_code']
            if 'psp_prefix' in proj._specific_template.keys():
                psp_prefix = proj._specific_template['psp_prefix']
            else:
                psp_prefix = ""

            if self._prof_gen:
                psp_code = self._prof_gen.annotate_computesum_rate(proj, psp_code)

            return psp_prefix, psp_code

        # Dictionary of keywords to transform the parsed equations
        ids = deepcopy(self._template_ids)

        # Use specialized code templates?
        if isinstance(proj.synapse_type, DefaultRateCodedSynapse) or \
           proj.synapse_type.description['psp']['eq']=="w*pre.r":

            simd_type = get_highest_available_simd_instructions_type(proj.net_id)

            # Does our current system support AVX?
            if simd_type is not None:

                try:
                    # The default weighted sum can be re-formulated for single weights
                    if proj._has_single_weight():
                        template = self._templates["vectorized_default_psp"][simd_type]["single_w"]
                    else:
                        template = self._templates["vectorized_default_psp"][simd_type]["multi_w"]

                    # the access to pre-synaptic firing depends on the delay
                    if proj.max_delay <= 1:
                        # no synaptic delay
                        ids.update({
                            'get_r': ids['pre_prefix']+"r.data()"
                        })
                        psp_code = template["sum"][ConfigManager().get('precision', self._net_id)] % ids

                        if self._prof_gen:
                            psp_code = self._prof_gen.annotate_computesum_rate(proj, psp_code)

                        return "", psp_code

                    elif proj.uniform_delay != -1 and proj.max_delay > 1:
                        # Uniform delay
                        ids.update({
                            'get_r': ids['pre_prefix']+"_delayed_r[delay-1].data()",
                        })
                        psp_code = template["sum"][ConfigManager().get('precision', self._net_id)] % ids

                        if self._prof_gen:
                            psp_code = self._prof_gen.annotate_computesum_rate(proj, psp_code)

                        return "", psp_code

                    else:
                        # HD (3rd June 2021): for non-uniform delays I doubt that it's worth the effort, so we proceed with
                        #                     general code generation
                        pass

                except KeyError:
                    # No fitting code found, so we fall back to normal code generation
                    # TODO: add internal error log, which key was missing?
                    Messages._debug("No SIMD implementation found, fallback to non-SIMD code")
                    template = ""

            else:
                # Other optimizations like loop unroll etc?
                if proj._storage_format == "bsr":
                    try:
                        # not yet implemented
                        if proj._has_single_weight():
                            raise KeyError

                        # Either user-selected or by ANNarchy
                        blockDim = proj._bsr_tile_size

                        # Loop unroll depends on the block-size
                        unrolled_template = template = self._templates["unrolled_default_psp"][blockDim]

                        # Check if we implemented a SIMD version
                        if simd_type in unrolled_template.keys():
                            template = unrolled_template[simd_type]['multi_w']['sum'][ConfigManager().get('precision', self._net_id)]
                        else:
                            template = unrolled_template['none']['multi_w']["sum"]

                        # the access to pre-synaptic firing depends on the delay
                        if proj.max_delay <= 1:
                            # no synaptic delay
                            ids.update({
                                'get_r': ids['pre_prefix']+"r.data()",
                            })

                            psp_code = template % ids

                            if self._prof_gen:
                                psp_code = self._prof_gen.annotate_computesum_rate(proj, psp_code)

                            return "", psp_code
                        else:
                            # HD (23th Nov 2021): the unrolled code templates for specific kernels is a highly
                            #                     experimental feature. I will implement the delay case if we
                            #                     are sure that we really use this.
                            pass

                    except KeyError:
                        # No fitting code found, so we fall back to normal code generation
                        # TODO: add internal error log, which key was missing?
                        Messages._debug("No SIMD implementation found, fallback to non-SIMD code")
                        template = ""

        # Choose the relevant summation template
        try:
            template = self._templates['rate_coded_sum']
        except:
            Messages._error("OpenMPGenerator: no template for storage_format = "+proj._storage_format+" configuration available")

        # Dictionary of keywords to transform the parsed equations
        ids = self._template_ids

        # Dependencies
        dependencies = list(set(proj.synapse_type.description['dependencies']['pre']))

        # Retrieve the PSP
        if not 'psp' in  proj.synapse_type.description.keys(): # default
            psp = """%(preprefix)s.r%(pre_index)s * w%(local_index)s;"""
        else: # custom psp
            psp = (proj.synapse_type.description['psp']['cpp'])

        # Special case where w is a single value
        if proj._has_single_weight():
            psp = re.sub(
                r'([^\w]+)w%\(local_index\)s',
                r'\1w',
                ' ' + psp
            )

        # Allow the use of global variables in psp (issue60)
        for var in dependencies:
            if var in proj.pre.neuron_type.description['global']:
                psp = psp.replace("%(pre_prefix)s"+var+"%(pre_index)s", "%(pre_prefix)s"+var+"%(global_index)s")

        # Delayed variables
        if isinstance(proj.pre, PopulationView):
            delayed_variables = proj.pre.population.delayed_variables
        else:
            delayed_variables = proj.pre.delayed_variables

        # Delays
        if proj.max_delay > 1: # There is non-zero delay
            if proj.uniform_delay == -1: # Non-uniform delays
                for var in delayed_variables:
                    if var in proj.pre.neuron_type.description['local']:
                        psp = psp.replace(
                            '%(pre_prefix)s'+var+'%(pre_index)s',
                            '%(pre_prefix)s_delayed_'+var+'%(delay_nu)s%(pre_index)s'
                        )
                    else:
                        Messages._print(proj.synapse_type.description['psp']['eq'])
                        Messages._error('The psp accesses a global variable with a non-uniform delay!')


            else: # Uniform delays
                for var in delayed_variables:
                    if var in proj.pre.neuron_type.description['local']:
                        psp = psp.replace(
                            '%(pre_prefix)s'+var+'%(pre_index)s',
                            '%(pre_prefix)s_delayed_'+var+'%(delay_u)s%(pre_index)s'
                        )
                    else:
                        psp = psp.replace(
                            '%(pre_prefix)s'+var+'%(global_index)s',
                            '%(pre_prefix)s_delayed_'+var+'%(delay_u)s'
                        )

        # OpenMP run modifiers
        schedule = "" if not 'psp_schedule' in proj._omp_config.keys() else proj._omp_config['psp_schedule']

        # Some formats have additional thread-local parameters
        if proj._storage_format=='lil':
            add_clause = 'private(sum)'
        elif proj._storage_format in ['ell', 'ellr']:
            add_clause = 'private(sum) firstprivate(maxnzr_)'
        else:
            add_clause = ''

        # In case of a uniform delay we preload the variable and then provide as
        # thread local argument to the worker threads
        pre_copy = ""
        first_privates = ""
        if proj.max_delay > 1 and proj.uniform_delay != -1:

            for var in dependencies:
                if var in proj.pre.neuron_type.description['local']:
                    pre_copy += "std::vector<%(float_prec)s> _pre_" + var + " = %(pre_prefix)s_delayed_" + var + "%(delay_u)s;"
                    psp = psp.replace(
                        '%(pre_prefix)s_delayed_'+var+'%(delay_u)s%(pre_index)s',
                        '_pre_'+var+'%(pre_index)s'
                    )
                    first_privates += '_pre_%(var)s, ' % {'var': var}

        # Special case for diagonal format
        if proj._storage_format == "dia":
            ids.update({'omp_simd': '#pragma omp for' if ConfigManager().get('disable_SIMD_SpMV', self._net_id) else '#pragma omp for simd'})

        # Finalize the psp with the correct ids
        psp = psp % ids
        pre_copy = pre_copy % ids

        ids.update({
            'pre_copy': pre_copy,
            'omp_code': '#pragma omp for',
            'omp_clause': add_clause,
            'omp_schedule': schedule,
            'psp': psp.replace(';', ''),
            'simd_len': str(4) if _check_precision('double', self._net_id) else str(8),
        })
        # Generate the code depending on the operation
        sum_code = template[proj.synapse_type.operation] % ids

        # Default variables needed in psp_code
        if proj._storage_format == "lil":
            psp_prefix = """
        %(idx_type)s nb_post, nb_pre;
        %(float_prec)s sum;""" % ids
        elif proj._storage_format == "ellr":
            psp_prefix = """
        %(idx_type)s rk_post, rk_pre;
        %(float_prec)s sum;""" % ids
        else:
            psp_prefix = ""

        # Finish the code
        final_code = """
        if (_transmission && %(post_prefix)s_active){
%(code)s
        } // active
        """ % {
            'post_prefix': ids['post_prefix'],
            'code': tabify(sum_code, 3),
        }

        if self._prof_gen:
            final_code = self._prof_gen.annotate_computesum_rate(proj, final_code)

        return psp_prefix, final_code

    def _computesum_spiking(self, proj, single_matrix):
        """
        Generate codes for spike propagation and pre-spike part of event-driven equations.

        Returns:

            * psp_prefix: set of variables needed in the kernel, positioned at the begin of
                          Projection::compute_psp() method.
            * code: computation body

        Specific templates:

            * psp_prefix and psp_code
        """
        if 'psp_prefix' in proj._specific_template.keys() and 'psp_code' in proj._specific_template.keys():
            psp_prefix = proj._specific_template['psp_prefix']
            psp_code = proj._specific_template['psp_code']

            if len(psp_code) > 0 and self._prof_gen:
                psp_code = self._prof_gen.annotate_computesum_spiking(proj, psp_code)

            return psp_prefix, psp_code

        # If the connectivity is stored as post_to_pre, we need to use the
        # inversed matrix view to acces the psp/weight vectors. As we
        # parallelize over pre-neurons, there is a chance of concurrent accesses
        # towards psp.
        #
        # Early implementations used atomics to protect, a clear performance
        # limiter. The user ilyasm proposed a solution using shared arrays and
        # a following reduction. Here we initialize the thread local array.
        if proj._storage_order == "post_to_pre":
            psp_prefix = ""
            if not proj.disable_omp: # TODO: are there other conditions?
                psp_prefix += """
#ifdef _OPENMP"""
                targets = [proj.target] if type(proj.target) == str else proj.target
                for target in targets:
                    psp_prefix += """
        std::vector< double > pop%(id)s_%(target)s_thr(pop%(id)s->get_size()*global_num_threads, 0.0);""" % { 'id': proj.post.id, 'target': target }
                psp_prefix += """
#endif
"""

            psp_prefix += """
        int nb_post;
        double sum;"""
        else:
            psp_prefix = """
        int nb_post;
        double sum;"""

        # Basic tags, dependent on storage format are assuming a feedforward
        # transmission.
        ids = deepcopy(self._template_ids)

        # The psp kernel sometimes use diverging indices
        # HD (10th April 2022): maybe I should remove this in future (TODO)
        if proj._storage_format == "lil":
            if ConfigManager().get('num_threads', self._net_id) == 1 or single_matrix:
                ids.update({
                    'pre_index': '[rk_j]',
                })
            else:
                ids.update({
                    'pre_index': '[rk_j]',
                })

        elif proj._storage_format == "csr":
            if proj._storage_order == "post_to_pre":
                if single_matrix:
                    ids.update({
                        'local_index': "[_inv_idx[syn]]",
                        'semiglobal_index': '[_row_idx[syn]]',
                        'global_index': '',
                        'pre_index': '[_pre]',
                        'post_index': '[_row_idx[syn]]',
                    })
                else:
                    ids.update({
                        'local_index': "[tid][_inv_idx[syn]]",
                        'semiglobal_index': '[_row_idx[syn]]',
                        'global_index': '',
                        'pre_index': '[_pre]',
                        'post_index': '[_row_idx[syn]]',
                    })

            else:
                if single_matrix:
                    ids.update({
                        'local_index': "[syn]",
                        'semiglobal_index': '[col_idx_[syn]]',
                        'global_index': '',
                        'pre_index': '[_pre]',
                        'post_index': '[col_idx_[syn]]',
                    })

                else:
                    ids.update({
                        'local_index': "[tid][syn]",
                        'semiglobal_index': '[col_idx_[syn]]',
                        'global_index': '',
                        'pre_index': '[_pre]',
                        'post_index': '[col_idx_[syn]]',
                    })

        elif proj._storage_format == "dense":
            if single_matrix:
                # Nothing to do
                pass
            else:
                raise NotImplementedError

        else:
            raise NotImplementedError

        # Determine the mode of synaptic transmission
        continous_transmission = False
        if 'psp' in  proj.synapse_type.description.keys(): # continous
            continous_transmission = True

        ####################################################
        # Event-driven summation of g_target
        ####################################################
        # Strings
        updated_variables_list = []
        g_target = ""
        g_target_code = ""

        # Analyse all elements of pre_spike
        for eq in proj.synapse_type.description['pre_spike']:
            # g_target is treated differently
            # Must be at the end of the equations
            if eq['name'] == 'g_target':
                # PSP form
                g_target = eq['cpp'].split('=')[1]
                # Operation (g_target is replaced by sum in 'cpp')
                operation = re.search(r'sum (.*?)=', eq['cpp']).group(1).strip() + "="
                # Check targets
                if isinstance(proj.target, str):
                    targets = [proj.target]
                else:
                    targets = proj.target
                g_target_code = ""
                for target in targets:
                    # Special case where w is a single value
                    if proj._has_single_weight():
                        g_target = re.sub(
                            r'([^\w]+)w%\(local_index\)s',
                            r'\1w',
                            g_target
                        )

                    target_dict = {
                        'post_prefix': ids['post_prefix'],
                        'target': target,
                        'g_target': g_target % ids,
                        'eq': eq['eq'],
                        'post_index': ids['post_index'],
                        'operation': operation
                    }

                    # access to post variable migth require atomic
                    # operation ( added later if needed )
                    if proj.max_delay > 1 and proj.uniform_delay == -1: # TODO: openMP is switched off for non uniform delays
                        g_target_code += """
            %(post_prefix)sg_%(target)s%(post_index)s %(operation)s %(g_target)s
"""% target_dict
                    elif proj.disable_omp or ConfigManager().get('num_threads', self._net_id) == 1:
                        g_target_code += """
            %(post_prefix)sg_%(target)s%(post_index)s %(operation)s %(g_target)s
"""% target_dict
                    else:
                        raise NotImplementedError

                    # Determine bounds
                    for key, val in eq['bounds'].items():
                        if not key in ['min', 'max']:
                            continue
                        try:
                            value = str(float(val))
                        except: # TODO: more complex operations
                            value = val % ids

                        g_target_code += """
            if (%(post_prefix)sg_%(target)s%(post_index)s %(op)s %(val)s)
                %(post_prefix)sg_%(target)s%(post_index)s = %(val)s;
""" % {'post_prefix': ids['post_prefix'], 'target': target, 'post_index': ids['post_index'], 'op': "<" if key == 'min' else '>', 'val': value}

            else:
                # process equations in pre_spike which
                # are not 'g_target'

                condition = ""
                # Check conditions to update the variable
                if eq['name'] == 'w': # Surround it by the learning flag
                    condition = "_plasticity" # Plasticity can be disabled

                if 'unless_post' in eq['flags']: # Flags avoids pre-spike evaluation when post fires at the same time
                    simultaneous = "%(pre_prefix)slast_spike[pre_rank[i][j]] != %(post_prefix)slast_spike[post_rank[i]]" % {'id_post': proj.post.id, 'id_pre': proj.pre.id}
                    if condition == "":
                        condition = simultaneous
                    else:
                        condition += "&&(" + simultaneous + ")"

                eq_dict = {
                    'eq': eq['eq'],
                    'cpp': eq['cpp'] % ids,
                    'bounds': get_bounds(eq) % ids,
                    'condition': condition
                }

                # Generate the code, either with or without coundition
                if condition != "":
                    updated_variables_list += """
// unless_post can prevent evaluation of presynaptic variables
if (%(condition)s) {
    // %(eq)s
    %(cpp)s
    %(bounds)s
}
""" % eq_dict
                else: # Normal synaptic variable
                    updated_variables_list += """
// %(eq)s
%(cpp)s
%(bounds)s""" % eq_dict


        # Generate the default post-conductance increase
        # default g_target += w
        if not continous_transmission and g_target == "":
            # Check targets
            if isinstance(proj.target, str):
                targets = [proj.target]
            else:
                targets = proj.target

            g_target_code = ""
            for target in targets:
                g_target_code += """
            // Increase the post-synaptic conductance g_target += w
            %(post_prefix)sg_%(target)s%(post_index)s += w%(local_index)s;
"""

        # Special case where w is a single value
        if proj._has_single_weight():
            g_target_code = re.sub(
                r'([^\w]+)w%\(local_index\)s',
                r'\1w',
                g_target_code
            )

        # finalize g_target_code
        g_target_code = g_target_code % ids

        # Event-driven integration of synaptic variables
        has_exact = False
        event_driven_code = ''
        for var in proj.synapse_type.description['variables']:
            if var['method'] == 'event-driven':
                has_exact = True
                event_driven_code += """
            // %(eq)s
            %(exact)s
""" % {'eq': var['eq'], 'exact': var['cpp'].replace('(t)', '(t-1)') % ids}
        if has_exact:
            event_driven_code += """
            // Update the last event for the synapse
            _last_event%(local_index)s = t;
""" % ids

        # Generate code for pre-spike variables
        pre_code = ""
        if len(updated_variables_list) > 0:
            for var in updated_variables_list:
                pre_code += var
            pre_code = tabify(pre_code, 3)

            # Special case where w is a single value
            if proj._has_single_weight():
                pre_code = re.sub(
                    r'([^\w]+)w\[i\]\[j\]',
                    r'\1w',
                    pre_code
                )

        # Choose correct template based on connectivity
        # and take delays into account if any
        pre_array = ""
        if proj.max_delay > 1:
            if proj.uniform_delay == -1: # Non-uniform delays
                Messages._warning('Variable delays for spiking networks is experimental and slow...')
                template = self._templates['spiking_sum_variable_delay']
            else: # Uniform delays
                template = self._templates['spiking_sum_fixed_delay'][proj._parallel_pattern]
                pre_array = "%(pre_prefix)s_delayed_spike[delay-1]" % ids
        else:
            pre_array = "%(pre_prefix)sspiked" % ids
            # For dense matrices we have two code templates either using with or without mask[idx]
            if g_target_code != "" and pre_code == "" and proj._storage_format in ["dense"]:
                template = self._templates['spiking_sum_fixed_delay_only_psp'][proj._parallel_pattern]
            else:
                template = self._templates['spiking_sum_fixed_delay'][proj._parallel_pattern]

        # sanity check
        if template == None:
            raise Messages.CodeGeneratorException("\tproj{}: no template available (Configuration: format={}, order={}, single_matrix={}, pattern={})".format(proj.id, proj._storage_format, proj._storage_order, single_matrix, proj._parallel_pattern))

        # Axonal spike events
        spiked_array_fusion_code = ""
        if proj.synapse_type.pre_axon_spike:
            spiked_array_fusion_code = """
    std::vector<int> tmp_spiked = %(pre_array)s;
    if (_axon_transmission) {
        tmp_spiked.insert( tmp_spiked.end(), %(pre_prefix)saxonal.begin(), %(pre_prefix)saxonal.end() );
    }
""" % {'pre_prefix': ids['pre_prefix'], 'pre_array': pre_array}

            pre_array = "tmp_spiked"

        # Generate the whole code block
        code = ""
        if g_target_code != "" or pre_code != "":
            ids.update({
                'pre_array': pre_array,
                'pre_event': pre_code,
                'g_target': g_target_code,
                'event_driven': event_driven_code,
                'spiked_array_fusion': spiked_array_fusion_code
            })
            code = template % ids

        # Add tabs
        code = tabify(code, 2)

        ####################################################
        # Not even-driven summation of psp: like rate-coded
        ####################################################
        if 'psp' in  proj.synapse_type.description.keys(): # not event-based
            # Compute it as if it were rate-coded
            _, psp_code = self._computesum_rate(proj, single_matrix)
            
            psp_prefix = tabify("%(float_prec)s sum; int nb_pre, nb_post;" % {'float_prec': ConfigManager().get('precision', self._net_id)}, 2)

            # Change _sum_target into g_target (TODO: handling of PopulationViews???)
            psp_code = psp_code.replace(
                '%(post_prefix)s_sum_%(target)s' % ids,
                '%(post_prefix)sg_%(target)s' % ids
            )

            # Add it to the main code
            code += """
        // PSP-based summation"""
            code += psp_code

        # Annotate code
        if self._prof_gen:
            code = self._prof_gen.annotate_computesum_spiking(proj, code)

        return psp_prefix, code

    def _header_structural_plasticity(self, proj):
        """
        Generate extension code for C header_struct: variable declaration, add and remove synapses.

        Templates:

            structural_plasticity: 'header_struct' field contains all relevant code templates

        """
        header_tpl = self._templates['structural_plasticity']['header_struct']

        code = ""
        # Pruning defined in the synapse
        if 'pruning' in proj.synapse_type.description.keys():
            code += header_tpl['pruning']

        # Creating defined in the synapse
        if 'creating' in proj.synapse_type.description.keys():
            code += header_tpl['creating']

        # Retrieve the names of extra attributes
        extra_args = ""
        extra_args_acc = ""
        extra_args_vec_decl = ""
        extra_args_vec_acc = ""
        add_var_code = ""
        add_var_remove = ""
        for var in proj.synapse_type.description['parameters'] + proj.synapse_type.description['variables']:
            if not var['name'] in ['w', 'delay'] and  var['name'] in proj.synapse_type.description['local']:

                if not isinstance(proj.init[var['name']], (int, float, bool)):
                    init = var['init']
                else:
                    init = proj.init[var['name']]
                extra_args += ', ' + var['ctype'] + ' _' +  var['name'] +'='+str(init)
                extra_args_acc += ', _' +  var['name']
                extra_args_vec_decl += ', std::vector<'+ var['ctype'] +'> _' + var['name']
                extra_args_vec_acc += ', _' + var['name'] + '[new_item]'
                add_var_code += ' '*8 + var['name'] + '[post_idx].insert('+var['name']+'[post_idx].begin() + pos, _' + var['name'] + ');\n'
                add_var_remove += ' '*8 + var['name'] + '[post_idx].erase(' + var['name'] + '[post_idx].begin() + pos);\n'

        # Delays
        delay_code = ""
        delay_remove= ""
        if proj.max_delay > 1 and proj.uniform_delay == -1:
            delay_code = ' '*8 + "delay[post].insert(delay[post_idx].begin() + pos, _delay);"
            delay_remove = ' '*8 + "delay[post].erase(delay[post_idx].begin() + pos);"

        # Spiking networks must update the inv_pre_rank array
        spiking_addcode = "" if proj.synapse_type.type == 'rate' else header_tpl['spiking_addcode']
        spiking_removecode = "" if proj.synapse_type.type == 'rate' else header_tpl['spiking_removecode']

        # Random distributions
        rd_addcode = ""
        rd_removecode = ""
        for rd in proj.synapse_type.description['random_distributions']:
            rd_addcode += """
        %(name)s[post].insert(%(name)s[post].begin() + idx, 0.0);
""" % {'name': rd['name']}

            rd_removecode += """
        %(name)s[post].erase(%(name)s[post].begin() + idx);
""" % {'name': rd['name']}

        # For spiking models, we need to rebuild the backward view, if synapses are removed/added
        inverse_connectivity_code = '' if proj.synapse_type.type == 'rate' else header_tpl['spiking_rebuild_backwardview']

        # Generate the code
        code += header_tpl['header'] % {
            'float_prec': ConfigManager().get('precision', self._net_id),
            'extra_args': extra_args, 'extra_args_acc': extra_args_acc,
            'extra_args_vec_decl': extra_args_vec_decl, 'extra_args_vec_acc': extra_args_vec_acc,
            'delay_code': delay_code, 'delay_remove': delay_remove,
            'add_code': add_var_code, 'add_remove': add_var_remove,
            'spike_add': spiking_addcode, 'spike_remove': spiking_removecode,
            'rd_add': rd_addcode, 'rd_remove': rd_removecode,
            'inverse_connectivity_rebuild': inverse_connectivity_code
        }

        return code

    def _init_random_distributions(self, proj):
        # Is it a specific population?
        if 'init_rng' in proj._specific_template.keys():
            return proj._specific_template['init_rng']

        code = ""
        for rd in proj.synapse_type.description['random_distributions']:
            rd_init = rd['definition'] % self._template_ids
            code += """    %(rd_name)s = std::vector< std::vector<double> >(post_rank.size(), std::vector<double>());
    for(int i=0; i<post_rank.size(); i++){
        %(rd_name)s[i] = std::vector<double>(pre_rank[i].size(), 0.0);
    }
    dist_%(rd_name)s = %(rd_init)s;
""" % {'rd_name': rd['name'], 'rd_init': rd_init}
        return code

    def _local_functions(self, proj):
        " Local functions "
        local_func = ""
        if len(proj.synapse_type.description['functions']) > 0:
            local_func += """
    // Local functions
"""
            for func in proj.synapse_type.description['functions']:
                local_func += ' '*4 + func['cpp'] + '\n'

        return local_func

    def _post_event(self, proj, single_matrix):
        """
        Generates the code for the post-synaptic updates of event-driven learning rules.
        """
        if proj.synapse_type.type == "rate":
            return "", ""

        if proj.synapse_type.description['post_spike'] == []:
            return "", ""

        # Get basic template ids and update if necessary.
        ids = deepcopy(self._template_ids)
        if proj._storage_format == "lil":
            ids.update({
                'post_index': '[rk_post]',
            })

        elif proj._storage_format == "csr":
            if proj._storage_order == "post_to_pre":
                if single_matrix:
                    ids.update({
                        'semiglobal_index': '[*it]',
                        'pre_index': '[_col_idx[j]]',
                        'post_index': '[rk_post]',
                    })
                else:
                    ids.update({
                        'local_index': "[tid][j]",
                        'semiglobal_index': '[tid][*it]',
                        'pre_index': '[_col_idx[j]]',
                        'post_index': '[rk_post]',
                    })

            else:
                if single_matrix:
                    ids.update({
                        'local_index': "[inv_idx_[j]]",
                        'semiglobal_index': '[*it]',
                        'global_index': '',
                        'pre_index': '[row_idx_[j]]',
                        'post_index': '[rk_post]',
                    })
                else:
                    ids.update({
                        'local_index': "[tid][inv_idx_[j]]",
                        'semiglobal_index': "[tid][*it]",
                        'global_index': "",
                        'pre_index': "[row_idx_[j]]",
                        'post_index': "[rk_post]",
                    })

        elif proj._storage_format == "dense":
            if proj._storage_order == "post_to_pre":
                if single_matrix:
                    pass    # currently working on
                else:
                    raise NotImplementedError
            else:
                if single_matrix:
                    pass    # currently working on
                else:
                    raise NotImplementedError

        else:
            # Sanity: this should not reached
            raise NotImplementedError

        # Definition of format-specific local variables.
        if proj._storage_format == "lil":
            post_event_prefix = ""
        elif proj._storage_format == "csr":
            post_event_prefix = """
        int rk_post;
        std::vector<int>::iterator it;
        """
        elif proj._storage_format == "dense":
            post_event_prefix = """
        int rk_post;
"""
        else:
            raise NotImplementedError

        # Event-driven integration
        has_event_driven = False
        for var in proj.synapse_type.description['variables']:
            if var['method'] == 'event-driven':
                has_event_driven = True

        # Generate event-driven code
        event_driven_code = ""
        if has_event_driven:
            for var in proj.synapse_type.description['variables']:
                if var['method'] == 'event-driven':
                    event_driven_code += '// ' + var['eq'] + '\n'
                    event_driven_code += var['cpp'] % ids + '\n'
            event_driven_code += """
// Update the last event for the synapse
_last_event%(local_index)s = t;
""" % ids

            event_driven_code = tabify(event_driven_code, 3)

        # Gather the equations
        post_code = ""
        for eq in proj.synapse_type.description['post_spike']:
            post_code += '// ' + eq['eq'] + '\n'
            if eq['name'] == 'w':
                post_code += "if(_plasticity)\n"
            post_code += eq['cpp'] % ids + '\n'
            post_code += get_bounds(eq) % ids + '\n'
        post_code = tabify(post_code, 3)

        # Generate the code block
        try:
            ids.update({
                'post_event': post_code,
                'event_driven': event_driven_code
            })
            code = self._templates['post_event'] % ids

        except KeyError:
            # Template does not exist
            Messages._error("No template for spiking neurons post event (format = {}, order = {}, and single_matrix = {})".format(proj._storage_format, proj._storage_order, single_matrix))

        # Annotate code
        if self._prof_gen:
            code = self._prof_gen.annotate_post_event(proj, code)

        return post_event_prefix, tabify(code, 2)

    def _update_random_distributions(self, proj):
        """
        Step-wise update of random distributed variables which may appear as local (per synapse),
        semiglobal (per dendrite) or global (one value per projection).

        TODO: implement parallel RNG (currently one thread is responsible for draw the random numbers)
        """
        # Is it a specific projection?
        if 'update_rng' in proj._specific_template.keys():
            return proj._specific_template['update_rng']

        if len(proj.synapse_type.description['random_distributions']) == 0:
            return ""

        global_code=""
        semiglobal_code=""
        local_code=""

        for rd in proj.synapse_type.description['random_distributions']:
            if rd['name'] in proj.synapse_type.description["local"]:
                local_code += self._templates["rng_update"][rd["locality"]] % {'rd_name': rd['name']}

        return tabify(self._templates["rng_update"]["template"] % {
            'global_rng': global_code,
            'semiglobal_rng': semiglobal_code,
            'local_rng': local_code,
            'idx_type': self._template_ids['idx_type']
        }, 2)


    def _update_synapse(self, proj, single_matrix):
        """Updates the continuously changed variables of the projection."""

        # Global variables
        global_eq = generate_equation_code(proj.synapse_type.description, 'global', 'proj', padding=2, wrap_w="_plasticity")

        # Code layout
        off = 1 if not single_matrix else 0 # fix tabs for sliced matrix
        off = 1 if proj._storage_format=="dense" else 0 # fix tabs for dense matrix

        # Semiglobal variables
        semiglobal_eq = generate_equation_code(proj.synapse_type.description, 'semiglobal', 'proj', padding=2+off, wrap_w="_plasticity")

        # Local variables
        local_eq = generate_equation_code(proj.synapse_type.description, 'local', 'proj', padding=3+off, wrap_w="_plasticity")

        # Skip generation if there are no equations
        if local_eq.strip() == '' and semiglobal_eq.strip() == '' and global_eq.strip() == '':
            return "", ""

        # Gather pre-loop declaration (dt/tau for ODEs)
        pre_code = ""
        for var in proj.synapse_type.description['variables']:
            if 'pre_loop' in var.keys() and len(var['pre_loop']) > 0:
                pre_code += var['ctype'] + ' ' + var['pre_loop']['name'] + ' = ' + var['pre_loop']['value'] + ';\n'

        if len(pre_code) > 0:
            pre_code = """
    // Updating the step sizes
""" + tabify(pre_code, 1)
            global_eq = pre_code + global_eq

        # adjust dt dependent on the _update_period, this covers only
        # the switch statements
        global_eq = re.sub(
            r'([^\w]+)dt([^\w]+)',
            r'\1_dt\2',
            global_eq
        )
        semiglobal_eq = re.sub(
            r'([^\w]+)dt([^\w]+)',
            r'\1_dt\2',
            semiglobal_eq
        )
        local_eq = re.sub(
            r'([^\w]+)dt([^\w]+)',
            r'\1_dt\2',
            local_eq
        )

        # Special case where w is a single value
        if proj._has_single_weight():
            local_eq = re.sub(
                r'([^\w]+)w%\(local_index\)s',
                r'\1w',
                ' ' + local_eq
            )
            global_eq = re.sub(
                r'([^\w]+)w%\(local_index\)s',
                r'\1w',
                ' ' + global_eq
            )

        # Dependencies
        dependencies = list(set(proj.synapse_type.description['dependencies']['pre']))

        # Take delays into account if any
        if proj.max_delay > 1:
            if proj.uniform_delay == -1: # Non-uniform delays
                for var in dependencies:
                    if var in proj.pre.neuron_type.description['local']:
                        local_eq = local_eq.replace(
                            '%(pre_prefix)s'+var+'%(pre_index)s',
                            '%(pre_prefix)s_delayed_'+var+'%(delay_nu)s%(pre_index)s'
                        )
                    else:
                        local_eq = local_eq.replace(
                            '%(pre_prefix)s'+var+'%(pre_index)s',
                            '%(pre_prefix)s_delayed_'+var+'%(delay_nu)s'
                        )
            else: # Uniform delays
                for var in dependencies:
                    if var in proj.pre.neuron_type.description['local']:
                        local_eq = local_eq.replace(
                            '%(pre_prefix)s'+var+'%(pre_index)s',
                            '%(pre_prefix)s_delayed_'+var+'%(delay_u)s%(pre_index)s'
                        )
                    else:
                        local_eq = local_eq.replace(
                            '%(pre_prefix)s'+var+'%(pre_index)s',
                            '%(pre_prefix)s_delayed_'+var+'%(delay_u)s'
                        )

        # Choose the template
        try:
            template = self._templates['update_variables']

        except KeyError:
            # either no template code at all, or no 'update_variables' field.
            Messages._error("No synaptic plasticity template found for format = " + proj._storage_format, ", order = " + proj._storage_order, " and single_matrix =", single_matrix)

        ids = deepcopy(self._template_ids)

        # Fill the code template
        if local_eq.strip() != "": # local synapses are updated
            ids.update({
                'global': global_eq % self._template_ids,
                'semiglobal': semiglobal_eq % self._template_ids,
                'local': local_eq % self._template_ids,
            })
            code = template['local'] % ids
        else: # Only global variables
            ids.update({
                'global': global_eq % self._template_ids,
                'semiglobal': semiglobal_eq % self._template_ids,
            })
            code = template['global'] % ids

        psp_prefix = """
        %(idx_type)s rk_post, rk_pre;
        %(float_prec)s _dt = dt * _update_period;""" % ids

        if self._prof_gen:
            code = self._prof_gen.annotate_update_synapse(proj, code)

        # Return the code block
        return psp_prefix, tabify(code, 2)

    def _update_max_delay(self, proj):
        "When the maximum delay of a non-uniform spiking projection changes, the ring buffer for delyed spikes must be updated."

        if proj.synapse_type.type == 'rate':
            return "", ""

        if proj.uniform_delay >= 0:
            return "", ""

        if proj._no_split_matrix:
            update_delay_code = """
        // No need to do anything if the new max delay is smaller than the old one
        if(d <= max_delay)
            return;

        // Update delays
        int prev_max = max_delay;
        max_delay = d;
        int add_steps = d - prev_max;

    #ifdef _DEBUG
        std::cout << "Delayed arrays was " << _delayed_spikes.size() << std::endl;
    #endif

        // Insert as many empty vectors as need at the current pointer position
        _delayed_spikes.insert(_delayed_spikes.begin() + idx_delay, add_steps, std::vector< std::vector< int > >(post_rank.size(), std::vector< int >() ));

        // The delay index has to be updated
        idx_delay = (idx_delay + add_steps) % max_delay;
"""
        else:
            update_delay_code = """
        // No need to do anything if the new max delay is smaller than the old one
        if(d <= max_delay)
            return;

        // Update delays
        int prev_max = max_delay;
        max_delay = d;
        int add_steps = d - prev_max;

    #ifdef _DEBUG
        std::cout << "Delayed arrays was " << std::endl;
        for (int tid = 0; tid < global_num_threads; tid++)
            std::cout << _delayed_spikes.size() << " for thread " << tid << std::endl;
    #endif

        // Insert as many empty vectors as need at the current pointer position
        for (int tid = 0; tid < global_num_threads; tid++) {
            _delayed_spikes[tid].insert(_delayed_spikes[tid].begin() + idx_delay, add_steps, std::vector< std::vector< int > >(sub_matrices_[tid]->post_rank.size(), std::vector< int >() ));
        }

        // The delay index has to be updated
        idx_delay = (idx_delay + add_steps) % max_delay;

    #ifdef _DEBUG
        std::cout << "Delayed arrays is now " << _delayed_spikes.size() << std::endl;
        std::cout << "Idx " << idx_delay << std::endl;
        for(int i = 0; i< max_delay; i++)
            std::cout << _delayed_spikes[i][0].size() << std::endl;
    #endif
"""

        reset_ring_buffer_code = self._templates['delay']['nonuniform_spiking']['reset'] % self._template_ids

        return update_delay_code, reset_ring_buffer_code



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector< std::vector<%(type)s> > %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable< %(type)s, std::vector<%(type)s> >(%(init)s);
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = std::vector<%(type)s>( post_ranks_.size(), %(init)s);
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();       
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'semiglobal': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'global': ""
}

delay = {
    # A single value for all synapses
    'uniform': {
        'declare': """
    // Uniform delay
    int delay;

    int get_delay() { return delay; }
    int get_dendrite_delay(int idx) { return delay; }
    void set_delay(int delay) { this->delay = delay; }
""",
        'init': """
    delay = delays[0][0];
"""
    },
    # An individual value for each synapse
    'nonuniform_rate_coded': {
        'declare': """
    std::vector<int> delay;
    int max_delay;

    std::vector<std::vector<int>> get_delay() { return get_matrix_variable_all<int>(delay); }
    void set_delay(std::vector<std::vector<int>> value) { update_matrix_variable_all<int>(delay, value); }
    std::vector<int> get_dendrite_delay(int lil_idx) { return get_matrix_variable_row<int>(delay, lil_idx); }
    int get_max_delay() { return max_delay; }
    void set_max_delay() { this->max_delay = max_delay; }
""",
        'init': """
    delay = init_matrix_variable<int>(1);
    update_matrix_variable_all<int>(delay, delays);
""",
        'reset': ""
    },
    # An individual value for each synapse and a
    # buffer for spike events    
    'nonuniform_spiking': {
        'declare': """
    std::vector<int> delay;
    int max_delay;
    int idx_delay;
    std::vector< std::vector< std::vector< int > > > _delayed_spikes;

    std::vector<std::vector<int>> get_delay() { return get_matrix_variable_all<int>(delay); }
    void set_delay(std::vector<std::vector<int>> value) { update_matrix_variable_all<int>(delay, value); }
    std::vector<int> get_dendrite_delay(int lil_idx) { return get_matrix_variable_row<int>(delay, lil_idx); }
    int get_max_delay() { return max_delay; }
    void set_max_delay() { this->max_delay = max_delay; }
""",
        'init': """
    delay = init_matrix_variable<int>(1);
    update_variable_all<int>(delay, delays);

    idx_delay = 0;
    max_delay = %(pre_prefix)smax_delay;
""",
        'reset': """
        while(!_delayed_spikes.empty()) {
            auto elem = _delayed_spikes.back();
            elem.clear();
            _delayed_spikes.pop_back();
        }

        idx_delay = 0;
        max_delay =  %(pre_prefix)smax_delay ;
        _delayed_spikes = std::vector< std::vector< std::vector< int > > >(max_delay, std::vector< std::vector< int > >(post_rank.size(), std::vector< int >()) );        
"""
    }    
}

event_driven = {
    'declare': """
    std::vector<std::vector<long>> _last_event;
""",
    'cpp_init': """
    _last_event = init_matrix_variable<long, std::vector<long>>(-10000);
""",
    'pyx_struct': """
        vector[vector[long]] _last_event
""",
}

update_variables = {
    'local': """
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L) ){
    %(global)s

    auto row_ptr = sub_matrices_[tid]->row_ptr();
    auto col_idx = sub_matrices_[tid]->row_indices();
    auto post_ranks = sub_matrices_[tid]->get_post_rank();
    %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_ranks.size());

    for(int i = 0; i < nb_post; i++){
        rk_post = post_ranks[i];
    %(semiglobal)s
        for(int j = row_ptr[rk_post]; j < row_ptr[rk_post+1]; j++){
            rk_pre = col_idx[j];
    %(local)s
        }
    }
}
""",
        'global': """
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L)) {
    %(global)s

    auto post_ranks = sub_matrices_[tid]->get_post_rank();
    %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_ranks.size());

    for(%(idx_type)s i = 0; i < nb_post; i++){
        rk_post = post_ranks[i];
    %(semiglobal)s
    }
}
"""
}

spiking_summation_fixed_delay_outer_loop = """// Event-based summation
if (_transmission && %(post_prefix)s_active) {
    auto _col_ptr = sub_matrices_[tid]->col_ptr();
    auto _row_idx = sub_matrices_[tid]->row_indices();
    auto _inv_idx = sub_matrices_[tid]->inverse_indices();

    for( int _idx = 0; _idx < %(pre_array)s.size(); _idx++) {
        int _pre = %(pre_array)s[_idx];

        // Iterate over connected post neurons
        for (int syn = _col_ptr[_pre]; syn < _col_ptr[_pre + 1]; syn++) {
            %(event_driven)s
            %(g_target)s
            %(pre_event)s
        }
    }
} // active
"""

spiking_post_event = """
// w as CSR
auto row_ptr = sub_matrices_[tid]->row_ptr();
auto _col_idx = sub_matrices_[tid]->column_indices();

if(_transmission && %(post_prefix)s_active){
    #pragma omp for
    for(int _idx_i = 0; _idx_i < %(post_prefix)sspiked.size(); _idx_i++){
        // Rank of the postsynaptic neuron which fired
        rk_post = %(post_prefix)sspiked[_idx_i];

        // Iterate over all synapse to this neuron
        for (int j = row_ptr[rk_post]; j < row_ptr[rk_post+1]; j++) {

%(event_driven)s
%(post_event)s
        }
    }
}
"""

conn_templates = {
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'delay': delay,
    'event_driven': event_driven,

    # operations
    'rate_coded_sum': None,
    'vectorized_default_psp': None,
    'spiking_sum_fixed_delay': {
        # HD: we use the same template, as the split variant always parallize across outer-loop.
        'inner_loop': spiking_summation_fixed_delay_outer_loop,
        'outer_loop': spiking_summation_fixed_delay_outer_loop
    },
    'spiking_sum_variable_delay': None,
    'update_variables': update_variables,
    'post_event': spiking_post_event
}

conn_ids = {
    'local_index': '[tid][j]',
    'semiglobal_index': '[i]',
    'global_index': '',
    'pre_index': '[col_idx[j]]',
    'post_index': '[post_ranks_[i]]',
    'delay_nu' : '[delay[j]-1]', # non-uniform delay
    'delay_u' : '[delay-1]' # uniform delay
}



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages

from .Equation import Equation
from .ParserTemplate import create_local_dict, user_functions

import sympy as sp

import re


class CoupledEquations(Equation):
    """
    Special equation solver when several equations are coupled and use the midpoint or implicit numerical methods.
    """

    def __init__(self, description, variables, net_id):
        # Global description
        self.description = description

        # List of equations to parse
        self.variables = variables

        # Net ID
        self.net_id = net_id

        # Build the list of expressions
        self.expression_list = {}
        for var in self.variables:
            self.expression_list[var['name']] = var['transformed_eq']

        # List of variable names to parse
        self.names = self.expression_list.keys()

        # Get attributes of the neuron/synapse
        self.local_attributes = self.description['local']
        self.semiglobal_attributes = self.description['semiglobal']
        self.global_attributes = self.description['global']
        self.untouched = variables[0]['untouched'] # the same for all eqs
        self.local_functions = [func['name'] for func in self.description['functions']]

        # Copy the default dictionary of built-in symbols or functions
        self.local_dict = create_local_dict(
            self.local_attributes, 
            self.semiglobal_attributes, 
            self.global_attributes, 
            self.untouched) # in ParserTemplate

        # Copy the list of built-in functions
        self.user_functions = user_functions.copy()
        # Add each user-defined function to avoid "not supported in C"
        for var in self.local_functions: 
            self.user_functions[var] = var

    def parse(self):
        "Main method called after creating the object."
        
        # Check if the numerical method is the same for all ODEs
        methods = []
        for var in self.variables:
            methods.append(var['method'])
        if len(list(set(methods))) > 1: # mixture of methods
            Messages._print(methods)
            Messages._error('Cannot mix different numerical methods when solving a coupled system of equations.')
            
        else:
            method = methods[0]

        if method == 'implicit' or method == 'semiimplicit':
            return self.solve_implicit(self.expression_list)
        elif method == 'midpoint': 
            return self.solve_midpoint(self.expression_list)
        elif method == 'rk4':
            return self.solve_rk4(self.expression_list)

    def solve_implicit(self, expression_list):
        "Implicit method"

        equations = {}
        new_vars = {}

        # Pre-processing to replace the gradient
        for name, expression in self.expression_list.items():
            # transform the expression to suppress =
            if '=' in expression:
                expression = expression.replace('=', '- (')
                expression += ')'
            # Suppress spaces to extract dvar/dt
            expression = expression.replace(' ', '')
            # Transform the gradient into a difference TODO: more robust...
            expression = expression.replace('d'+name, '_t_gradient_')
            expression_list[name] = expression

        # replace the variables by their future value
        for name, expression in expression_list.items():
            for n in self.names:
                expression = re.sub(r'([^\w]+)'+n+r'([^\w]+)', r'\1_'+n+r'\2', expression)
            expression = expression.replace('_t_gradient_', '(_'+name+' - '+name+')')
            expression_list[name] = expression

            new_var = sp.Symbol('_'+name)
            self.local_dict['_'+name] = new_var
            new_vars[new_var] = name

        for name, expression in expression_list.items():
            analysed = self.parse_expression(
                expression,
                local_dict = self.local_dict
            )
            equations[name] = analysed

        try:
            solution = sp.solve(list(equations.values()), list(new_vars.keys()))
        except Exception as e:
            Messages._print(expression_list)
            Messages._error('The multiple ODEs can not be solved together using the implicit Euler method.')     

        for var, sol in solution.items():
            # Simplify the solution
            sol  = sp.collect( sol, self.local_dict['dt'])

            # C++ code
            try:
                code = sp.ccode(sol, strict=False) 
            except:
                code = sp.ccode(sol) 


            # Generate the code
            cpp_eq = ConfigManager().get('precision', self.net_id) + ' _' + new_vars[var] + ' = ' + code + ';'
            switch = sp.ccode(self.local_dict[new_vars[var]] ) + ' = _' + new_vars[var] + ';'

            # Replace untouched variables with their original name
            for prev, new in sorted(list(self.untouched.items()), key = lambda key : len(key[0]), reverse=True):
                cpp_eq = re.sub(prev, new, cpp_eq)
                switch = re.sub(prev, new, switch)

            # Store the result
            for variable in self.variables:
                if variable['name'] == new_vars[var]:
                    variable['cpp'] = cpp_eq
                    variable['switch'] = switch

        return self.variables



    def solve_midpoint(self, expression_list):
        "Midpoint method"

        equations = {}
        evaluations = {}

        # Pre-processing to replace the gradient
        for name, expression in expression_list.items():
            
            # transform the expression to suppress =
            if '=' in expression:
                expression = expression.replace('=', '- (')
                expression += ')'
            
            # Suppress spaces to extract dvar/dt
            expression = expression.replace(' ', '')
            
            # Transform the gradient into a difference TODO: more robust..            
            expression = expression.replace('d'+name+'/dt', '_grad_var_'+name)
            new_var = sp.Symbol('_grad_var_'+name)
            self.local_dict['_grad_var_'+name] = new_var
            
            expression_list[name] = expression


        for name, expression in expression_list.items():

            analysed = self.parse_expression(expression,
                local_dict = self.local_dict
            )
            equations[name] = analysed
            evaluations[name] = sp.solve(analysed, self.local_dict['_grad_var_'+name])

        # Compute the k = f(x, t)
        ks = {}
        for name, evaluation in evaluations.items():
            ks[name] = ConfigManager().get('precision', self.net_id) + ' _k_' + name + ' = ' + self.c_code(evaluation[0]) + ';'

        # New dictionary replacing x by x+dt/2*k)
        tmp_dict = {}
        for name, val in self.local_dict.items():
            tmp_dict[name] = val
        for name, evaluation in evaluations.items():
            tmp_dict[name] = sp.Symbol('(' + sp.ccode(self.local_dict[name]) + ' + 0.5 * dt * _k_' + name + ' )')

        # Compute the new values _x_new = f(x + dt/2*_k)
        news = {}
        for name, expression in expression_list.items():
            tmp_analysed = self.parse_expression(expression,
                local_dict = tmp_dict
            )
            solved = sp.solve(tmp_analysed, self.local_dict['_grad_var_'+name])
            news[name] = ConfigManager().get('precision', self.net_id) + ' _' + name + ' = ' + self.c_code(solved[0]) + ';'

        # Compute the switches
        switches = {}
        for name, expression in expression_list.items():
            switches[name] = sp.ccode(self.local_dict[name]) + ' += dt * _' + name + ' ;'

        # Store the generated code in the variables
        for name in self.names:
            k = ks[name]
            n = news[name]
            switch = switches[name]

            # Replace untouched variables with their original name
            for prev, new in sorted(list(self.untouched.items()), key = lambda key : len(key[0]), reverse=True):
                k = re.sub(prev, new, k)
                n = re.sub(prev, new, n)
                switch = re.sub(prev, new, switch)

            # Store the result
            for variable in self.variables:
                if variable['name'] == name:
                    variable['cpp'] = [k, n]
                    variable['switch'] = switch
            
        return self.variables

    def solve_rk4(self, expression_list):
        "Runge-Kutta 4th order"

        expression_list = {}
        equations = {}
        evaluations = {}

        # Pre-processing to replace the gradient
        for name, expression in self.expression_list.items():
            # transform the expression to suppress =
            if '=' in expression:
                expression = expression.replace('=', '- (')
                expression += ')'
            # Suppress spaces to extract dvar/dt
            expression = expression.replace(' ', '')
            # Transform the gradient into a difference TODO: more robust...
            expression = expression.replace('d'+name+'/dt', '_gradient_'+name)
            self.local_dict['_gradient_'+name] = sp.Symbol('_gradient_'+name)
            expression_list[name] = expression

        for name, expression in expression_list.items():
            analysed = self.parse_expression(expression,
                local_dict = self.local_dict
            )
            equations[name] = analysed
            evaluations[name] = sp.solve(analysed, self.local_dict['_gradient_'+name])

        # Compute the k1 = f(x, t)
        k1_dict = {}
        for name, evaluation in evaluations.items():
            k1_dict[name] = ConfigManager().get('precision', self.net_id) + ' _k1_' + name + ' = ' + self.c_code(evaluation[0]) + ';'

        # New dictionary replacing x by x+dt/2*k1)
        k2_dict = {}
        tmp_dict_k2 = {}
        for name, val in self.local_dict.items():
            tmp_dict_k2[name] = val
        for name, evaluation in evaluations.items():
            tmp_dict_k2[name] = sp.Symbol('(' + sp.ccode(self.local_dict[name]) + ' + 0.5 * dt * _k1_' + name + ' )')

        # Compute the values _k2_x = f(x + dt/2*_k1)
        for name, expression in expression_list.items():
            tmp_analysed = self.parse_expression(expression,
                local_dict = tmp_dict_k2
            )

            solved = sp.solve(tmp_analysed, self.local_dict['_gradient_'+name])
            k2_dict[name] = ConfigManager().get('precision', self.net_id) + ' _k2_' + name + ' = ' + self.c_code(solved[0]) + ';'

        # New dictionary replacing x by x+dt/2*k2)
        k3_dict = {}
        tmp_dict_k3 = {}
        for name, val in self.local_dict.items():
            tmp_dict_k3[name] = val
        for name, evaluation in evaluations.items():
            tmp_dict_k3[name] = sp.Symbol('(' + sp.ccode(self.local_dict[name]) + ' + 0.5 * dt *_k2_' + name + ' )')

        # Compute the values _k3_x = f(x + dt/2*_k2_x)
        for name, expression in expression_list.items():
            tmp_analysed = self.parse_expression(expression,
                local_dict = tmp_dict_k3
            )

            solved = sp.solve(tmp_analysed, self.local_dict['_gradient_'+name])
            k3_dict[name] = ConfigManager().get('precision', self.net_id) + ' _k3_' + name + ' = ' + self.c_code(solved[0]) + ';'

        # New dictionary replacing x by x+dt*k3)
        k4_dict = {}
        tmp_dict_k4 = {}
        for name, val in self.local_dict.items():
            tmp_dict_k4[name] = val
        for name, evaluation in evaluations.items():
            tmp_dict_k4[name] = sp.Symbol('(' + sp.ccode(self.local_dict[name]) + ' + dt*_k3_' + name + ' )')

        # Compute the values _k4_x = f(x + dt*_k3_x)
        for name, expression in expression_list.items():
            tmp_analysed = self.parse_expression(expression,
                local_dict = tmp_dict_k4
            )

            solved = sp.solve(tmp_analysed, self.local_dict['_gradient_'+name])
            k4_dict[name] = ConfigManager().get('precision', self.net_id) + ' _k4_' + name + ' = ' + self.c_code(solved[0]) + ';'

        # accumulate _k1 - _k4 within the switch step
        dt_code = "dt/6.0f" if ConfigManager().get('precision', self.net_id) == 'float' else "dt/6.0"
        switches = {}
        for name, expression in expression_list.items():
            switches[name] = sp.ccode(self.local_dict[name]) + ' += '+dt_code+' * (_k1_'+name+' + (_k2_'+name+' + _k2_'+name+') + (_k3_'+name+' + _k3_'+name+') + _k4_'+name+');'

        # Store the generated code in the variables
        for name in self.names:
            k1 = k1_dict[name]
            k2 = k2_dict[name]
            k3 = k3_dict[name]
            k4 = k4_dict[name]
            switch = switches[name]

            # Replace untouched variables with their original name
            for prev, new in sorted(list(self.untouched.items()), key = lambda key : len(key[0]), reverse=True):
                k1 = re.sub(prev, new, k1)
                k2 = re.sub(prev, new, k2)
                k3 = re.sub(prev, new, k3)
                k4 = re.sub(prev, new, k4)
                switch = re.sub(prev, new, switch)

            # Store the result
            for variable in self.variables:
                if variable['name'] == name:
                    variable['cpp'] = [k1, k2, k3, k4]
                    variable['switch'] = switch

        return self.variables


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Random import *
from ANNarchy.core.Parameters import Parameter, Variable, Creating, Pruning
from ANNarchy.parser.Equation import Equation
from ANNarchy.parser.Function import FunctionParser
from ANNarchy.parser.StringManipulation import *
from ANNarchy.parser.ITE import *
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.intern.GlobalObjects import GlobalObjectManager
from ANNarchy.intern import Messages

import re


def extract_randomdist(description, net_id):
    " Extracts RandomDistribution objects from all variables"
    rk_rand = 0
    random_objects = []
    for variable in description['variables']:
        # Equation
        eq = variable['eq']
        # Dependencies
        dependencies = []
        # Search for all distributions
        for dist in available_distributions:
            matches = re.findall(r'(?P<pre>[^\w.])'+dist+r'\(([^()]+)\)', eq)
            if matches == ' ':
                continue
            for l, v in matches:

                # Check the arguments
                arguments = v.split(',')

                # Check the number of provided arguments
                if len(arguments) < distributions_arguments[dist]:
                    Messages._print(eq)
                    Messages._error('The distribution ' + dist + ' requires ' + str(distributions_arguments[dist]) + 'parameters')
                elif len(arguments) > distributions_arguments[dist]:
                    Messages._print(eq)
                    Messages._error('Too many parameters provided to the distribution ' + dist)

                # Process the arguments
                processed_arguments = ""
                for idx in range(len(arguments)):
                    try:
                        arg = float(arguments[idx])
                    except: # A global parameter
                        if arguments[idx].strip() in description['global']:
                            arg = arguments[idx].strip() + "%(global_index)s"
                            dependencies.append(arguments[idx].strip())
                        else:
                            Messages._error(arguments[idx] + ' is not a global parameter of the neuron/synapse. It can not be used as an argument to the random distribution ' + dist + '(' + v + ')')

                    processed_arguments += str(arg)
                    if idx != len(arguments)-1: # not the last one
                        processed_arguments += ', '

                definition = distributions_equivalents[dist] + '(' + processed_arguments + ')'

                # Store its definition
                desc = {
                    'name': 'rand_' + str(rk_rand),
                    'dist': dist,
                    'definition': definition,
                    'args': processed_arguments,
                    'template': distributions_equivalents[dist],
                    'locality': variable['locality'],
                    'ctype': 'double',
                    'dependencies': dependencies
                }
                rk_rand += 1
                random_objects.append(desc)

                # Replace its definition by its temporary name
                # Problem: when one uses twice the same RD in a single equation (perverse...)
                eq = eq.replace(dist+'('+v+')', desc['name'])
                # Add the new variable to the vocabulary
                description['attributes'].append(desc['name'])
                if variable['name'] in description['local']:
                    description['local'].append(desc['name'])
                elif variable['name'] in description['semiglobal']:
                    description['semiglobal'].append(desc['name'])
                else: # Why not on a population-wide variable?
                    description['global'].append(desc['name'])

        variable['transformed_eq'] = eq

    return random_objects

def extract_globalops_neuron(name, eq, description, net_id):
    """ Replaces global operations (mean(r), etc)  with arbitrary names and
    returns a dictionary of changes.
    """
    untouched = {}
    globs = []
    # Global ops
    glop_names = ['min', 'max', 'mean', 'norm1', 'norm2']
    for op in glop_names:
        matches = re.findall(r'([^\w]*)' + op + r'\(([\s\w]*)\)', eq)
        for pre, var in matches:
            if var.strip() in description['local']:
                globs.append({'function': op, 'variable': var.strip()})
                oldname = op + '(' + var + ')'
                newname = '_' + op + '_' + var.strip()
                eq = eq.replace(oldname, newname)
                untouched[newname] = '_' + op + '_' + var.strip()
            else:
                Messages._print(eq)
                Messages._error('There is no local attribute '+var+'.')

    return eq, untouched, globs

def extract_globalops_synapse(name, eq, desc, net_id):
    """
    Replaces global operations (mean(pre.r), etc)  with arbitrary names and
    returns a dictionary of changes.
    """
    untouched = {}
    globs = {'pre' : [],
             'post' : [] }
    glop_names = ['min', 'max', 'mean', 'norm1', 'norm2']
    origin_eq = eq  # temporary copy only for error message

    for op in glop_names:
        pre_matches = re.findall(r'([^\w.])' + op + r'\(\s*pre\.([\w]+)\s*\)', eq)
        post_matches = re.findall(r'([^\w.])' + op + r'\(\s*post\.([\w]+)\s*\)', eq)
        proj_matches = re.findall(r'([^\w.])' + op + r'\(\s*([\w]+)\s*\)', eq)

        for pre, var in pre_matches:
            globs['pre'].append({'function': op, 'variable': var.strip()})
            newname =  '__pre_' + op + '_' + var.strip()
            eq = re.sub(op+r'\(\s*pre\.([\w]+)\s*\)', newname, eq)
            untouched[newname] = '%(pre_prefix)s_' + op + '_' + var

        for pre, var in post_matches:
            globs['post'].append({'function': op, 'variable': var.strip()})
            newname = '__post_' + op + '_' + var.strip()
            eq = re.sub(op+r'\(\s*post\.([\w]+)\s*\)', newname, eq)
            untouched[newname] = '%(post_prefix)s_' + op + '_' + var

        for _, var in proj_matches:
            Messages._error("Detected global operation '"+op+"' on synaptic variable '"+var+"' in equation '"+origin_eq+"' which is not implemented.")

    return eq, untouched, globs

def extract_prepost(name, eq, description, net_id):
    " Replaces pre.var and post.var with arbitrary names and returns a dictionary of changes."

    dependencies = {'pre': [], 'post': []}

    pre_matches = re.findall(r'pre\.([\w]+)', eq)
    post_matches = re.findall(r'post\.([\w]+)', eq)

    untouched = {}
    # Replace all pre.* occurences with a temporary variable
    for var in list(set(pre_matches)):
        if var == 'sum': # pre.sum(exc)
            def idx_target(val):
                target = val.group(1).strip()
                if target == '':
                    Messages._print(eq)
                    Messages._error('pre.sum() requires one argument.')

                rep = '_pre_sum_' + target.strip()
                dependencies['pre'].append('sum('+target+')')
                untouched[rep] = '%(pre_prefix)s_sum_' +target+ '%(pre_index)s'
                return rep

            eq = re.sub(r'pre\.sum\(([\s\w]+)\)', idx_target, eq)
        else:
            dependencies['pre'].append(var)
            eq = re.sub(r"pre." + var + r"([^_\w]+)", "_pre_" + var + r"__\g<1>", eq + " ")
            # eq = eq.replace(target, ' _pre_'+var)
            untouched['_pre_'+var+'__'] = '%(pre_prefix)s' + var + '%(pre_index)s'

    # Replace all post.* occurences with a temporary variable
    for var in list(set(post_matches)):
        if var == 'sum': # post.sum(exc)
            def idx_target(val):
                target = val.group(1).strip()
                if target == '':
                    Messages._print(eq)
                    Messages._error('post.sum() requires one argument.')

                dependencies['post'].append('sum('+target+')')
                rep = '_post_sum_' + target.strip()
                untouched[rep] = '%(post_prefix)s_sum_' + target + '%(post_index)s'
                return rep
            eq = re.sub(r'post\.sum\(([\s\w]+)\)', idx_target, eq)
        else:
            dependencies['post'].append(var)
            eq = re.sub(r"post." + var + r"([^_\w]+)", r"_post_" + var + r"__\g<1>", eq + " ")
            # eq = eq.replace(target, ' _post_'+var+'__')
            untouched['_post_'+var+'__'] = '%(post_prefix)s' + var +'%(post_index)s'

    return eq, untouched, dependencies


def extract_parameters(description, extra_values={}, object_type="neuron", net_id=0):
    """
    Extracts all variable information from a multiline description or a dictionary.
    """
    result = []

    if isinstance(description, (dict,)):

        for key, val in description.items():
            
            # Locality is determined by localparam/globalparam. The default is global
            if isinstance(val, (Parameter,)):
                # Get fields
                value = val.value
                locality = val.locality
                if locality in ['population', 'projection']: # for people used to this
                    locality = 'global'
                if locality in ['postsynaptic']:
                    locality = 'semiglobal'
                if not locality in ['global', 'semiglobal', 'local']:
                    Messages._error(f"Parameter {key}: the locality must be in ['global', 'semiglobal', 'local'].")
                
                ctype = val.type
                # Possibility to give the built-in type instead of the string
                if ctype == float:
                    ctype = 'float'
                if ctype == int:
                    ctype = 'int'
                if ctype == bool:
                    ctype = 'bool'
                if not ctype in ['float', 'int', 'bool']:
                    Messages._error(f"Parameter {key}: the data type must be in ['float', 'int', 'bool'].")

                # Recreate the string flags for compatibility
                if locality == 'global':
                    flags = 'population' if object_type == 'neuron' else 'projection'
                elif locality == 'semiglobal':
                    flags = 'postsynaptic'
                else:
                    flags = ""
                if ctype in ['int', 'bool']:
                    flags += ", " + ctype
            else: 
                locality = 'global'
                value = val
                ctype = 'float'
                flags = 'population' if object_type == 'neuron' else 'projection'

            # ctype 
            ctype = ConfigManager().get('precision', net_id) if ctype == 'float' else ctype

            # Store the result
            desc = {'name': key,
                    'locality': locality,
                    'eq': key + " = " + str(value),
                    'bounds': {},
                    'flags' : flags,
                    'ctype' : ctype,
                    'init' : value,
                    }
            result.append(desc)
    
    elif isinstance(description, (str, list, )):
    
        # Split the multilines into individual lines
        if isinstance(description, (str,)):
            description = prepare_string(description)

        # Analyse all variables
        for definition in description:
            # Check if there are flags after the : symbol
            equation, constraint = split_equation(definition)
            
            # Extract the name of the variable
            name = extract_name(equation)
            if name in ['_undefined', ""]:
                Messages._error("Definition can not be analysed: " + equation)

            # Process constraint
            bounds, flags, ctype, init = extract_boundsflags(constraint, equation, extra_values, net_id=net_id)

            # Determine locality
            for f in ['population', 'postsynaptic', 'projection']:
                if f in flags:
                    if f == 'postsynaptic':
                        locality = 'semiglobal'
                    else:
                        locality = 'global'
                    break
            else:
                locality = 'local'

            # Store the result
            desc = {'name': name,
                    'locality': locality,
                    'eq': equation,
                    'bounds': bounds,
                    'flags' : flags,
                    'ctype' : ctype,
                    'init' : init,
                    }
            result.append(desc)
    else:
        Messages._error(f"parameters {description} should either be a dictionary or a multi-string.")

    return result

def extract_variables(description, object_type="neuron", net_id=0):
    """
    Extracts all variable information from a multiline description.

    object_type is either 'neuron' or 'synapse', to get the right locality keywords. (yes, we map global to population/projection back to global...)
    """
    
    
    # If it is a list, convert it to the old multistring. Future versions should reverse the process.
    if isinstance(description, (Variable,)):
        
        description = description._to_string(object_type)
    
    elif isinstance(description, (list,)):

        string_description = ""

        for eq in description:

            # If it is a string, just append it
            if isinstance(eq, (str,)):
                string_description += "\n" + eq
            elif isinstance(eq, (Variable,)):
                string_description += "\n" + eq._to_string(object_type)

        description = string_description
    elif isinstance (description, (str,)):
        pass
    else:
        Messages._error("equations must be either a string or a list of strings/variables.")
    

    # Split the multilines into individual lines
    variable_list = process_equations(description)
    
    # Analyse all variables
    result = []
    for definition in variable_list:

        # Retrieve the name, equation and constraints for the variable
        equation = definition['eq']
        constraint = definition['constraint']
        name = definition['name']
        if name == '_undefined':
            Messages._error('The variable', name, 'can not be analysed.')

        # Check the validity of the equation
        check_equation(equation)

        # Process constraint
        bounds, flags, ctype, init = extract_boundsflags(constraint, net_id=net_id)

        # Determine locality
        for f in ['population', 'postsynaptic', 'projection']:
            if f in flags:
                if f == 'postsynaptic':
                    locality = 'semiglobal'
                else:
                    locality = 'global'
                break
        else:
            locality = 'local'

        # Store the result
        desc = {'name': name,
                'locality': locality,
                'eq': equation,
                'bounds': bounds,
                'flags' : flags,
                'ctype' : ctype,
                'init' : init }
        result.append(desc)

    return result

def extract_boundsflags(constraint, equation="", extra_values={}, net_id=0):
    """
    Implementation note on boolean values:

    The selected C++ data type to represent boolean values differs for CPU and GPU. The reason
    is that a std::vector<bool> is a specialized implementation within the C++ standard template
    library (see https://en.cppreference.com/w/cpp/container/vector_bool).

    According to the reference, this vector "does not necessarily store its elements
    as a contiguous array". Therefore, the access to the raw data using the
    std::vector::data() method is not allowed which would be required to implement
    a host-to-device / device-to-host memory transfer.

    Therefore, we use the following rule: `bool` for single-thread and openMP and `char`
    for CUDA.
    """

    # Process the flags if any
    bounds, flags = extract_flags(constraint)

    # Get the type of the variable (float/int/bool).
    if 'int' in flags:
        ctype = 'int'
    elif 'bool' in flags:
        ctype = 'bool' if _check_paradigm("openmp", net_id) else 'char'
    else:
        ctype = ConfigManager().get('precision', net_id)

    # Detect the random distributions
    random_pattern = r'(' + '|'.join(available_distributions) + r')\s*\([^)]*\)'

    # Get the init value if declared
    if 'init' in bounds.keys(): # Variables: explicitely set in init=xx
        init = bounds['init']
        if ctype == 'bool':
            if init in ['false', 'False', '0']:
                init = False
            elif init in ['true', 'True', '1']:
                init = True
        elif isinstance(init, str) and re.search(random_pattern, init) is not None:
            init = eval(init)
        else:
            try:
                if ctype == 'int':
                    init = int(init)
                else:
                    init = float(init)
            except:
                init = str(init) # Check later whether it is a valid parameter name

    elif '=' in equation: # Parameters: the value is in the equation
        init = equation.split('=')[1].strip()

        # Boolean
        if init in ['false', 'False']:
            init = False
            ctype = 'bool' if _check_paradigm("openmp", net_id) else 'char'
        elif init in ['true', 'True']:
            init = True
            ctype = 'bool' if _check_paradigm("openmp", net_id) else 'char'
        # Extra-args (obsolete)
        elif init.strip().startswith("'"):
            var = init.replace("'","")
            init = extra_values[var]
        # Integers
        elif ctype == 'int':
            try:
                init = eval('int(' + init + ')')
            except:
                Messages._print(equation)
                Messages._error('The value of the parameter is not an integer.')
        # Floats
        else:
            try:
                init = eval('float(' + init + ')')
            except:
                Messages._print(equation)
                Messages._error('The value of the parameter is not a float.')

    else: # Default = 0 according to ctype
        if ctype == 'bool':
            init = False
        elif ctype == 'int':
            init = 0
        elif ctype == 'double' or ctype == 'float':
            init = 0.0

    return bounds, flags, ctype, init

def extract_functions(description, net_id, local_global=False):
    """ Extracts all functions from a multiline description."""

    if not description:
        return []

    # Split the multilines into individual lines
    function_list = process_equations(description)

    # Process each function
    functions = []
    for f in function_list:
        eq = f['eq']
        var_name, content = eq.split('=', 1)
        # Extract the name of the function
        func_name = var_name.split('(', 1)[0].strip()
        # Extract the arguments
        arguments = (var_name.split('(', 1)[1].split(')')[0]).split(',')
        arguments = [arg.strip() for arg in arguments]

        # Check the function name is not reserved by Sympy
        from inspect import getmembers
        import sympy
        functions_list = [o[0] for o in getmembers(sympy)]
        if func_name in functions_list:
            Messages._error('The function name', func_name, 'is reserved by sympy. Use another one.')

        # Extract their types
        types = f['constraint']
        if types == '':
            return_type = ConfigManager().get('precision', net_id)
            arg_types = [ConfigManager().get('precision', net_id) for a in arguments]
        else:
            types = types.split(',')
            return_type = types[0].strip()
            arg_types = [arg.strip() for arg in types[1:]]
        if not len(arg_types) == len(arguments):
            Messages._error('You must specify exactly the types of return value and arguments in ' + eq)

        arg_line = ""
        for i in range(len(arguments)):
            arg_line += arg_types[i] + " " + arguments[i]
            if not i == len(arguments) -1:
                arg_line += ', '

        # Process the content
        eq2, condition = extract_ite('', content, {'attributes': [], 'local':[], 'global': [], 'variables': [], 'parameters': []}, split=False)
        if condition == []:
            parser = FunctionParser('', content, arguments)
            parsed_content = parser.parse()
        else:
            parsed_content, deps = translate_ITE("", eq2, condition, arguments, {}, function=True)
            arguments = list(set(arguments)) # somehow the entries in arguments are doubled ... ( HD, 23.02.2017 )

        # Create the one-liner
        fdict = {'name': func_name, 'args': arguments, 'content': content, 'return_type': return_type, 'arg_types': arg_types, 'parsed_content': parsed_content, 'arg_line': arg_line}
        if not local_global: # local to a class
            oneliner = """%(return_type)s %(name)s (%(arg_line)s) {return %(parsed_content)s ;};
""" % fdict
        else: # global
            oneliner = """inline %(return_type)s %(name)s (%(arg_line)s) {return %(parsed_content)s ;};
""" % fdict
        fdict['cpp'] = oneliner
        functions.append(fdict)

    return functions


def get_attributes(parameters, variables, neuron):
    """ 
    Returns a list of all attributes names, plus the lists of local/global variables.
    """
    attributes = []; local_var = []; global_var = []; semiglobal_var = []

    for p in parameters + variables:
        attributes.append(p['name'])
        if neuron:
            if 'population' in p['flags']:
                global_var.append(p['name'])
            elif 'projection' in p['flags']:
                Messages._error('The attribute', p['name'], 'belongs to a neuron, the flag "projection" is forbidden.')
            elif 'postsynaptic' in p['flags']:
                Messages._error('The attribute', p['name'], 'belongs to a neuron, the flag "postsynaptic" is forbidden.')
            else:
                local_var.append(p['name'])
        else:
            if 'population' in p['flags']:
                Messages._error('The attribute', p['name'], 'belongs to a synapse, the flag "population" is forbidden.')
            elif 'projection' in p['flags']:
                global_var.append(p['name'])
            elif 'postsynaptic' in p['flags']:
                semiglobal_var.append(p['name'])
            else:
                local_var.append(p['name'])

    return attributes, local_var, global_var, semiglobal_var

def extract_targets(variables):
    targets = []
    for var in variables:
        # Rate-coded neurons
        code = re.findall(r'(?P<pre>[^\w.])sum\(\s*([^()]+)\s*\)', var['eq'])
        for l, t in code:
            targets.append(t.strip())
        # Special case for sum()
        if len(re.findall(r'([^\w.])sum\(\)', var['eq'])) > 0:
            targets.append('__all__')

        # Spiking neurons
        code = re.findall(r'([^\w.])g_([\w]+)', var['eq'])
        for l, t in code:
            targets.append(t.strip())

    return list(set(targets))


def convert_to_multistring(equations):
    """
    Converts a list of variables into a multistring that can be passed to process_equations()
    """
        
    # If it is a list, convert it to the old multistring. Future versions should reverse the process.
    if isinstance(equations, (Variable,)):
        
        equations = equations._to_string('neuron')
    
    elif isinstance(equations, (list,)):

        string_equations = ""

        for eq in equations:

            # If it is a string, just append it
            if isinstance(eq, (str,)):
                string_equations += "\n" + eq
            elif isinstance(eq, (Variable,)):
                string_equations += "\n" + eq._to_string('neuron')
        
        equations = string_equations

    elif isinstance (equations, (str,)):
        # Already a string, all good
        pass

    else:
        Messages._error("equations must be either a string or a list of strings/variables.")

    return equations

def extract_spike_variable(description, net_id):

    # Spike condition
    cond = prepare_string(description['raw_spike'])
    if len(cond) > 1:
        Messages._print(description['raw_spike'])
        Messages._error('The spike condition must be a single expression')

    translator = Equation('raw_spike_cond',
                            cond[0].strip(),
                            description)
    raw_spike_code = translator.parse()
    # Also store the variables used in the condition, as it may be needed for CUDA generation
    spike_code_dependencies = translator.dependencies()

    # Reset
    reset_desc = []
    if 'raw_reset' in description.keys() and description['raw_reset']:

        # Get equations
        equations = description['raw_reset']

        # Convert the equations to a multistring
        equations = convert_to_multistring(equations)
        
        # Process each line
        reset_desc = process_equations(equations)
        
        for var in reset_desc:
            translator = Equation(var['name'], var['eq'],
                                  description)
            var['cpp'] = translator.parse()
            var['dependencies'] = translator.dependencies()

    return { 'spike_cond': raw_spike_code,
             'spike_cond_dependencies': spike_code_dependencies,
             'spike_reset': reset_desc}

def extract_axon_spike_condition(description, net_id):
    """
    Extract the condition for emitting an axonal spike event. Further
    the reset after the event is returned.
    """
    if description['raw_axon_spike'] == None:
        return None

    cond = prepare_string(description['raw_axon_spike'])
    if len(cond) > 1:
        Messages._print(description['raw_axon_spike'])
        Messages._error('The spike condition must be a single expression')

    translator = Equation('raw_axon_spike_cond',
                            cond[0].strip(),
                            description)
    raw_spike_code = translator.parse()
    # Also store the variables used in the condition, as it may be needed for CUDA generation
    spike_code_dependencies = translator.dependencies()

    reset_desc = []
    if 'raw_reset' in description.keys() and description['raw_axon_reset']:

        # Convert the equations to a multistring
        equations = convert_to_multistring(equations)

        # Process the equations
        reset_desc = process_equations(description['raw_axon_reset'])
        for var in reset_desc:
            translator = Equation(var['name'], var['eq'],
                                  description)
            var['cpp'] = translator.parse()
            var['dependencies'] = translator.dependencies()

    return {
        'spike_cond': raw_spike_code,
        'spike_cond_dependencies': spike_code_dependencies,
        'spike_reset': reset_desc
    }

def extract_pre_spike_variable(description, net_id):

    # Get the equations
    equations = description['raw_pre_spike']
    if equations is None:
        return []

    # Convert the equations to a multistring
    equations = convert_to_multistring(equations)

    # For all variables influenced by a presynaptic spike
    pre_spike_var = []
    for var in process_equations(equations):
        # Get its name
        name = var['name']
        eq = var['eq']

        # Process the flags if any
        bounds, flags, ctype, init = extract_boundsflags(var['constraint'], net_id=net_id)

        # Extract if-then-else statements
        #eq, condition = extract_ite(name, raw_eq, description)

        # Append the result of analysis
        pre_spike_var.append( { 'name': name, 'eq': eq ,
                                'locality': 'local',
                                'bounds': bounds,
                                'flags':flags, 'ctype' : ctype,
                                'init' : init} )

    return pre_spike_var

def extract_post_spike_variable(description, net_id):
    
    # Get the equations
    equations = description['raw_post_spike']
    if not equations:
        return []

    # Convert the equations to a multistring
    equations = convert_to_multistring(equations)

    post_spike_var = []
    for var in process_equations(equations):
        # Get its name
        name = var['name']
        eq = var['eq']

        # Process the flags if any
        bounds, flags, ctype, init = extract_boundsflags(var['constraint'], net_id=net_id)

        # Extract if-then-else statements
        #eq, condition = extract_ite(name, raw_eq, description)

        post_spike_var.append( { 'name': name, 'eq': eq, 'raw_eq' : eq,
                                'locality': 'local',
                                'bounds': bounds, 'flags':flags, 'ctype' : ctype, 'init' : init} )

    return post_spike_var

def extract_axon_spike_variable(description, net_id):
    
    # Get the equations
    equations = description['raw_axon_spike']
    if not equations: 
        return []

    # Convert the equations to a multistring
    equations = convert_to_multistring(equations)

    # For all variables influenced by a presynaptic spike
    axon_spike_var = []
    for var in process_equations(equations):
        # Get its name
        name = var['name']
        eq = var['eq']

        # Process the flags if any
        bounds, flags, ctype, init = extract_boundsflags(var['constraint'], net_id=net_id)

        # Append the result of analysis
        axon_spike_var.append( {'name': name, 'eq': eq ,
                                'locality': 'local',
                                'bounds': bounds,
                                'flags':flags, 'ctype' : ctype,
                                'init' : init} )

    return axon_spike_var

def extract_stop_condition(pop, net_id=0):

    eq = pop['stop_condition']['eq']
    pop['stop_condition']['type'] = 'any'

    # Check the flags
    split = eq.split(':')
    if len(split) > 1: # flag given
        eq = split[0]
        flags = split[1].strip()
        split = flags.split(' ')
        for el in split:
            if el.strip() == 'all':
                pop['stop_condition']['type'] = 'all'

    # Convert the expression
    translator = Equation('stop_cond', eq,
                          pop,
                          type = 'cond')
    code = translator.parse()
    deps = translator.dependencies()

    pop['stop_condition']['cpp'] = '(' + code + ')'
    pop['stop_condition']['dependencies'] = deps

def extract_structural_plasticity(statement, description, net_id=0):

    # Extract flags
    if isinstance(statement, (str,)):
        try:
            eq, constraint = statement.rsplit(':', 1)
            bounds, flags = extract_flags(constraint)
        except:
            eq = statement.strip()
            bounds = {}
            flags = []

    elif isinstance(statement, (Creating)):
        eq = statement.equation
        proba = statement.proba
        w = statement.w
        d = statement.d
        bounds = {'proba': str(proba), 'w': str(w)}
        if d is not None: bounds['d'] = str(d)
        flags = []

    elif isinstance(statement, (Pruning)):
        eq = statement.equation
        proba = statement.proba
        bounds = {'proba': str(proba)}
        flags = []

    # Extract RD
    rd = None
    for dist in available_distributions:
        matches = re.findall(r'(?P<pre>[^\w.])' + dist + r'\(([^()]+)\)', eq)
        for l, v in matches:
            # Check the arguments
            arguments = v.split(',')
            # Check the number of provided arguments
            if len(arguments) < distributions_arguments[dist]:
                Messages._print(eq)
                Messages._error('The distribution ' + dist + ' requires ' + str(distributions_arguments[dist]) + 'parameters')
            elif len(arguments) > distributions_arguments[dist]:
                Messages._print(eq)
                Messages._error('Too many parameters provided to the distribution ' + dist)
            # Process the arguments
            processed_arguments = ""
            for idx in range(len(arguments)):
                try:
                    arg = float(arguments[idx])
                except: # A global parameter
                    Messages._print(eq)
                    Messages._error('Random distributions for creating/pruning synapses must use foxed values.')

                processed_arguments += str(arg)
                if idx != len(arguments)-1: # not the last one
                    processed_arguments += ', '
            definition = distributions_equivalents[dist] + '(' + processed_arguments + ')'

            # Store its definition
            if rd:
                Messages._print(eq)
                Messages._error('Only one random distribution per equation is allowed.')


            rd = {
                'name': 'rand_' + str(0) ,
                'origin': dist+'('+v+')',
                'dist': dist,
                'definition': definition,
                'args' : processed_arguments,
                'template': distributions_equivalents[dist]
            }

    if rd:
        eq = eq.replace(rd['origin'], 'rd(rng)')

    # Extract pre/post dependencies
    eq, untouched, dependencies = extract_prepost('test', eq, description, net_id=net_id)

    # Parse code
    translator = Equation('test', eq,
                          description,
                          method = 'cond',
                          untouched = {})

    code = translator.parse()
    deps = translator.dependencies()

    # Replace untouched variables with their original name
    for prev, new in sorted(list(untouched.items()), key = lambda key : len(key[0]), reverse=True):
        code = code.replace(prev, new)

    # Add new dependencies
    for dep in dependencies['pre']:
        description['dependencies']['pre'].append(dep)
    for dep in dependencies['post']:
        description['dependencies']['post'].append(dep)

    return {'eq': eq, 'cpp': code, 'bounds': bounds, 'flags': flags, 'rd': rd, 'dependencies': deps}


def find_method(variable):

    if 'implicit' in variable['flags']:
        method = 'implicit'
    elif 'semiimplicit' in variable['flags']:
        method = 'semiimplicit'
    elif 'exponential' in variable['flags']:
        method = 'exponential'
    elif 'midpoint' in variable['flags']:
        method = 'midpoint'
    elif 'runge-kutta4' in variable['flags']: # old name, backward compatibility
        method = 'rk4'
    elif 'rk4' in variable['flags']:
        method = 'rk4'
    elif 'explicit' in variable['flags']:
        method = 'explicit'
    elif 'exact' in variable['flags']:
        Messages._warning('The "exact" flag should now be replaced by "event-driven". It will stop being valid in a future release.')
        method = 'event-driven'
    elif 'event-driven' in variable['flags']:
        method = 'event-driven'
    else:
        method = get_global_config('method')
        if method == "runge-kutta4": # old name, backward compatibility
            method = 'rk4'

    return method

def check_equation(equation):
    "Makes a formal check on the equation (matching parentheses, etc)"
    # Matching parentheses
    if equation.count('(') != equation.count(')'):
        Messages._print(equation)
        Messages._error('The number of parentheses does not match.')



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Neuron import Neuron

#=====================================================ENCODING TECHNIQUES =========================================================================
#
#	The Intrinsically Bursting and Chattering type of encoding has implemented by referring to the paper "Simple Model of Spiking Neurons"
#	by Eugene M. Izhikevich, this is achieved by manipulating the parameters of Izhikevich Neuron and the nature of the neuron activity
#	of the these two technichs are as of Burst Coding.
#
#	The Phase Coding Oscillation type of encoidng has implemented by referring ot the paper "Deep neural networks with weighted spikes"
#	by Jaehyun Kim et al., this is achieved by writing the equation of the phase(equation 6) of the referred paper and multiplying phase
#	with the threshold (vt being constant) value to set the new dynamic thershold.
#

__all__ = ["CPN", "IB", "PSO"]

#====================================================Custom Poisson Neuron========================================================================

CPN = Neuron(
    parameters = """
        rates = 0.0
        mask_tau = 20.0
    """,
    equations = """
        p = (Uniform(0.0, 1.0) * 1000.0) / dt
        dmask/dt = -mask/mask_tau : init=0.0
    """,
    spike = "p < rates",
    reset="mask += 1.0/mask_tau",
    name="custom poisson neuron"
)

#====================================================Intrinsically Bursting========================================================================

IB = Neuron(
    parameters = """
        a = 0.02
        b = 0.2
        c = -55.0
        d = 4.0
        v_thresh = 30.0
        rates=0.0
        mask_tau = 20
    """,
    equations = """
        dv/dt = 0.04 * v^2 + 5.0 * v + 140.0 - u + rates : init = -55.0
        du/dt = a * (b*v - u) : init= -13.0
        dmask/dt = -mask/mask_tau : init=0.0
    """,
    spike = "v > v_thresh",
    reset = "v = c; u += d; mask += 1.0/mask_tau",
    name = "intrinsically bursting"
)

#=====================================================phase shift oscillation==========================================================================

PSO = Neuron(
    parameters = """
        k= 8
        vt = 1
        rates=0
    """,
    equations = """
        p= pow(2,(-1+(modulo(t-1,k))))
        vt_new=p*vt
        v = rates : init = 0
        dmask/dt = -mask/mask_tau : init=0.0
    """,
    spike = "v > vt_new",
    reset = "mask += 1.0/mask_tau",
    name = "phase shift oscillation"
)

#=====================================================================================================================================================
